from __future__ import division, print_function, absolute_import
import numpy as np


# Second pertubation Eq 36
def a2m(suma_a2, khs, xi, cte_a2m):
    a2 = khs*(1+xi)*suma_a2*cte_a2m
    return a2


def da2m_deta(suma_a2, dKhs, dXi, cte_a2m):

    khs, dkhs = dKhs
    xi, dx1 = dXi
    x1 = 1. + xi

    #sum1, dsum1 = np.matmul(da1sb, x0lambda)
    sum1, dsum1 = suma_a2
    # a2 = khs*x1*eps*c2*sum1/2.
    a2 = khs*x1*sum1*cte_a2m

    da2 = sum1*x1*dkhs + khs * x1 * dsum1 + khs * sum1 * dx1
    da2 *= cte_a2m
    return np.hstack([a2, da2])


def d2a2m_deta(suma_a2,  d2Khs, d2Xi, cte_a2m):

    khs, dkhs, d2khs = d2Khs
    xi, dx1, d2x1 = d2Xi
    x1 = 1. + xi

    sum1, dsum1, d2sum1 = suma_a2

    a2 = khs*x1*sum1*cte_a2m

    da2 = sum1*x1*dkhs + khs * x1 * dsum1 + khs * sum1 * dx1
    da2 *= cte_a2m

    d2a2 = d2khs * sum1 * x1 + d2x1 * sum1 * khs + d2sum1 * khs * x1
    d2a2 += 2 * dkhs * dsum1 * x1
    d2a2 += 2 * sum1 * dkhs * dx1
    d2a2 += 2 * khs * dsum1 * dx1
    d2a2 *= cte_a2m
    return np.hstack([a2, da2, d2a2])


def da2m_new_deta(suma_a2, dKhs, cte_a2m):

    khs, dkhs = dKhs
    # sum1, dsum1 = np.matmul(da1sb, x0lambda)
    sum1, dsum1 = suma_a2
    da2 = cte_a2m*(dkhs*sum1 + khs * dsum1)

    return da2


def d2a2m_new_deta(suma_a2, d2Khs, cte_a2m):

    khs, dkhs, d2khs = d2Khs
    # sum1, dsum1, d2sum1 = np.matmul(d2a1sb, x0lambda)
    sum1, dsum1, d2sum1 = suma_a2
    # aux = cte_a2m
    da2 = cte_a2m*(dkhs*sum1 + khs * dsum1)

    d2a2 = 2 * dkhs * dsum1 + sum1 * d2khs + d2sum1 * khs
    d2a2 *= cte_a2m

    return np.hstack([da2, d2a2])


def d3a2m_new_deta(suma_a2, d3Khs, cte_a2m):

    khs, dkhs, d2khs, d3khs = d3Khs
    # sum1, dsum1, d2sum1, d3sum1 = np.matmul(d3a1sb, x0lambda)
    sum1, dsum1, d2sum1, d3sum1 = suma_a2
    # aux = cte_a2m
    da2 = cte_a2m*(dkhs*sum1 + khs * dsum1)

    d2a2 = 2 * dkhs * dsum1 + sum1 * d2khs + d2sum1 * khs
    d2a2 *= cte_a2m

    d3a2 = 3 * dsum1 * d2khs + 3 * dkhs * d2sum1
    d3a2 += khs * d3sum1 + d3khs * sum1
    d3a2 *= cte_a2m
    return np.hstack([da2, d2a2, d3a2])



from __future__ import division, print_function, absolute_import
import numpy as np
from scipy.optimize import root
from .equilibriumresult import EquilibriumResult


def haz_objb(inc, T_P, vlle_type, model):

    X, W, Y, P_T = np.array_split(inc, 4)
    P_T = P_T[0]
    if vlle_type == 'T':
        P = P_T
        temp_aux = T_P
    elif vlle_type == 'P':
        T = P_T
        temp_aux = model.temperature_aux(T)
        P = T_P

    global vx, vw, vy
    global Xassx, Xassw, Xassy

    fugX, vx, Xassx = model.logfugef_aux(X, temp_aux, P, 'L', vx, Xassx)
    if np.any(np.isnan(fugX)) or np.isnan(vx):
        fugX, vx, Xassx = model.logfugef_aux(X, temp_aux, P, 'L')

    fugW, vw, Xassw = model.logfugef_aux(W, temp_aux, P, 'L', vw, Xassw)
    if np.any(np.isnan(fugW)) or np.isnan(vw):
        fugW, vw, Xassw = model.logfugef_aux(W, temp_aux, P, 'L')

    fugY, vy, Xassy = model.logfugef_aux(Y, temp_aux, P, 'V', vy, Xassy)
    if np.any(np.isnan(fugW)) or np.isnan(vw):
        fugY, vy, Xassy = model.logfugef_aux(Y, temp_aux, P, 'V')

    K1 = np.exp(fugX-fugY)
    K2 = np.exp(fugX-fugW)
    return np.hstack([K1*X-Y, K2*X-W, X.sum()-1, Y.sum()-1, W.sum()-1])


def vlleb(X0, W0, Y0, P_T, T_P, spec, model, v0=[None, None, None],
          Xass0=[None, None, None], full_output=False):
    '''
    Solves liquid liquid vapour equilibrium for binary mixtures.
    (T,P) -> (x,w,y)

    Parameters
    ----------

    X0 : array_like
        guess composition of phase 1
    W0 : array_like
        guess composition of phase 1
    Y0 : array_like
        guess composition of phase 2
    P_T : float
        absolute temperature [K] or pressure [Pa]
    T_P : floar
        absolute temperature [K] or pressure [Pa]
    spec: string
        'T' if T_P is temperature or 'P' if pressure.
    model : object
        created from mixture and saftvrmie function
    v0 : list, optional
        if supplied volume used as initial value to compute fugacities
    Xass0, list, optional
        if supplied used to solve the associtaion nonbonded sites fraction
    full_output: bool, optional
        wheter to outputs all calculation info

    Returns
    -------

    X : array_like
        liquid1 mole fraction vector
    W : array_like
        liquid2 mole fraction vector
    Y : array_like
        vapour mole fraction fector
    var: float
        temperature or pressure, depending of specification

    '''

    nc = model.nc

    if nc != 2:
        raise Exception('3 phase equilibria for binary mixtures')

    if len(X0) != nc or len(W0) != nc or len(Y0) != nc:
        raise Exception('Composition vector lenght must be equal to nc')

    global vx, vw, vy
    vx, vw, vy = v0

    global Xassx, Xassw, Xassy
    Xassx, Xassw, Xassy = Xass0

    if spec == 'T':
        temp_aux = model.temperature_aux(T_P)
        sol1 = root(haz_objb, np.hstack([X0, W0, Y0, P_T]),
                    args=(temp_aux, spec, model))
    elif spec == 'P':
        sol1 = root(haz_objb, np.hstack([X0, W0, Y0, P_T]),
                    args=(T_P, spec, model))
    else:
        raise Exception('Specification not known')

    error = np.linalg.norm(sol1.fun)
    nfev = sol1.nfev
    sol = sol1.x
    if np.any(sol < 0):
        raise Exception('negative Composition or T/P  founded')
    X, W, Y, var = np.array_split(sol, 4)
    var = var[0]
    if full_output:
        if spec == 'T':
            P = var
            T = T_P
        elif spec == 'P':
            T = var
            temp_aux = model.temperature_aux(T)
            P = T_P

        rhox, Xassx = model.density_aux(X, temp_aux, P, 'L', rho0=1./vx,
                                        Xass0=Xassx)
        rhow, Xassw = model.density_aux(W, temp_aux, P, 'L', rho0=1./vw,
                                        Xass0=Xassw)
        rhoy, Xassy = model.density_aux(Y, temp_aux, P, 'V', rho0=1./vy,
                                        Xass0=Xassy)
        vx = 1./rhox
        vw = 1./rhow
        vy = 1./rhoy

        inc = {'T': T, 'P': P, 'error': error, 'nfev': nfev,
               'X': X, 'vx': vx, 'Xassx': Xassx, 'statex': 'Liquid',
               'W': W, 'vw': vw, 'Xassw': Xassw, 'statew': 'Liquid',
               'Y': Y, 'vy': vy, 'Xassy': Xassy, 'statey': 'Vapor'}
        out = EquilibriumResult(inc)
        return out

    return X, W, Y, var



from __future__ import division, print_function, absolute_import
import numpy as np
from .monomer_aux import eta_eff, deta_eff,  d2eta_eff,  d3eta_eff


def a1s(eta, lam, cctes, eps):
    # a1s calculation Eq 39
    neff = eta_eff(eta, cctes)

    ghs = (1 - neff/2) / (1 - neff)**3
    a1vdw = -12 * eps * eta / (lam - 3)
    a1 = a1vdw * ghs
    return a1


def da1s(eta, lam, cctes, eps):

    neff, dneff = deta_eff(eta, cctes)

    # a1s calculation Eq 39
    ghs = (1 - neff/2) / (1 - neff)**3
    a1vdw = -12 * eps * eta / (lam - 3)
    a1 = a1vdw * ghs

    # da1s calculation
    ter1a1 = -2 + 3*neff - neff**2 + eta*(-5 + 2*neff)*dneff
    ter2a1 = 6*eps/((lam-3)*(1-neff)**4)
    da1 = ter1a1 * ter2a1

    return np.hstack([a1, da1])


def d2a1s(eta, lam, cctes, eps):
    neff, dneff, d2neff = d2eta_eff(eta, cctes)

    # a1s calculation Eq 39
    ghs = (1 - neff/2) / (1 - neff)**3
    a1vdw = -12 * eps * eta / (lam - 3)
    a1 = a1vdw * ghs

    # da1s calculation
    ter1a1 = -2 + 3*neff - neff**2 + eta*(-5 + 2*neff)*dneff
    ter2a1 = 6*eps/((lam-3)*(1-neff)**4)
    da1 = ter1a1 * ter2a1

    # d2a1s calculation
    ter1a2 = 2. * dneff * (5. - 7*neff + 2*neff**2 - 3*eta * (-3 + neff)*dneff)
    ter1a2 += eta*(-1 + neff)*(-5+2*neff)*d2neff
    ter2a2 = 6*eps/((lam-3)*(-1+neff)**5)
    d2a1 = ter1a2 * ter2a2
    return np.hstack([a1, da1, d2a1])


def d3a1s(eta, lam, cctes, eps):

    neff, dneff, d2neff, d3neff = d3eta_eff(eta, cctes)

    # a1s calculation Eq 39
    ghs = (1 - neff/2) / (1 - neff)**3
    a1vdw = -12 * eps * eta / (lam - 3)
    a1 = a1vdw * ghs

    # da1s calculation
    ter1a1 = -2 + 3*neff - neff**2 + eta*(-5 + 2*neff)*dneff
    ter2a1 = 6*eps/((lam-3)*(1-neff)**4)
    da1 = ter1a1 * ter2a1

    # d2a1s calculation
    ter1a2 = 2. * dneff * (5. - 7*neff+2*neff**2 - 3*eta * (-3 + neff)*dneff)
    ter1a2 += eta*(-1 + neff)*(-5+2*neff)*d2neff
    ter2a2 = 6*eps/((lam-3)*(-1+neff)**5)
    d2a1 = ter1a2 * ter2a2

    # d3a1s calculation
    ter1a3 = -18. * (-3 + neff) * (-1 + neff) * dneff**2
    ter1a3 += 12. * eta * (-7 + 2*neff)*dneff**3
    ter1a3 += - 18 * eta*(-3 + neff) * (-1 + neff) * dneff * d2neff
    ter1a3 += (-1 + neff)**2 * (-5 + 2 * neff) * (3*d2neff + eta * d3neff)
    ter2a3 = 6*eps/((lam-3)*(-1+neff)**6)
    d3a1 = ter1a3 * ter2a3
    return np.hstack([a1, da1, d2a1, d3a1])



from __future__ import division, print_function, absolute_import
import numpy as np


# Equation (62) Paper 2014
def g2mca(xhi00, khs, xs_m, da2new, suma_g2, eps_ii, a1vdw_cteii):

    g2 = 3 * da2new / eps_ii
    g2 -= khs * suma_g2 / xhi00
    g2 /= xs_m
    g2 /= - a1vdw_cteii
    return g2


def dg2mca_dxhi00(xhi00, khs, dkhs, xs_m, d2a2new, dsuma_g2, eps_ii,
                  a1vdw_cteii):

    sum1, dsum1 = dsuma_g2
    g2 = 3. * np.asarray(d2a2new) / eps_ii

    g2[0] -= khs * sum1 / xhi00
    g2[1] += khs * sum1 / xhi00**2
    g2[1] -= (sum1 * dkhs + khs * dsum1)/xhi00
    g2 /= xs_m
    g2 /= - a1vdw_cteii
    return g2


def d2g2mca_dxhi00(xhi00, khs, dkhs, d2khs, xs_m, d3a2new, d2suma_g2,
                   eps_ii, a1vdw_cteii):

    sum1, dsum1, d2sum1 = d2suma_g2

    g2 = 3. * np.asarray(d3a2new) / eps_ii

    aux1 = khs * sum1 / xhi00
    aux2 = aux1 / xhi00
    aux3 = (sum1 * dkhs + khs * dsum1)/xhi00

    g2[0] -= aux1
    g2[1] += aux2
    g2[1] -= aux3
    g2[2] += 2*aux3/xhi00
    g2[2] -= 2*aux2/xhi00
    g2[2] -= (2*dkhs*dsum1 + sum1*d2khs + khs*d2sum1)/xhi00
    g2 /= xs_m
    g2 /= - a1vdw_cteii

    return g2



from __future__ import division, print_function, absolute_import
import numpy as np
from .monomer_aux import Xi, dXi_dxhi00, d2Xi_dxhi00


def a2(xs, khs, xhixm, a2ij, epsij, f1, f2, f3):

    xi = Xi(xhixm, f1, f2, f3)
    a2 = a2ij * khs * (1 + xi) * epsij/2
    a = np.matmul(np.matmul(a2, xs), xs)

    return a


def da2_dxhi00(xs, khs, dkhs, xhixm, dxhixm_dxhi00, da2ij, epsij, f1, f2, f3):

    xi, dxi = dXi_dxhi00(xhixm, dxhixm_dxhi00, f1, f2, f3)
    ctes = epsij / 2
    sum1, dsum1 = da2ij

    a2 = sum1 * khs * (1 + xi) * ctes
    da2 = sum1 * (1+xi) * dkhs + khs * (1 + xi) * dsum1 + khs * sum1 * dxi
    da2 *= ctes

    da = np.array([a2, da2])
    a = np.matmul(np.matmul(da, xs), xs)
    return a


def d2a2_dxhi00(xs, khs, dkhs, d2khs, xhixm, dxhixm_dxhi00, d2a2ij, epsij,
                f1, f2, f3):

    xi, dxi, d2xi = d2Xi_dxhi00(xhixm, dxhixm_dxhi00, f1, f2, f3)

    ctes = epsij / 2
    sum1, dsum1, d2sum1 = d2a2ij

    a2 = sum1 * khs * (1 + xi) * ctes

    da2 = sum1 * (1+xi) * dkhs + khs * (1 + xi) * dsum1 + khs * sum1 * dxi
    da2 *= ctes

    d2a2 = sum1 * (1 + xi) * d2khs
    d2a2 += sum1 * khs * d2xi
    d2a2 += d2sum1 * (1 + xi) * khs
    d2a2 += 2 * dkhs * ((1 + xi) * dsum1 + sum1 * dxi)
    d2a2 += 2 * khs * dsum1 * dxi
    d2a2 *= ctes

    da = np.array([a2, da2, d2a2])
    a = np.matmul(np.matmul(da, xs), xs)
    return a



from __future__ import division, print_function, absolute_import
import numpy as np
from scipy.optimize import fsolve, root
from .multiflash import multiflash
from .equilibriumresult import EquilibriumResult
from warnings import warn


def haz_objb(inc, T_P, type, model, index, equilibrium):

    X0 = inc[:-1].reshape(3, 2)
    P_T = inc[-1]

    if type == 'T':
        P = P_T
        temp_aux = T_P
    elif type == 'P':
        T = P_T
        temp_aux = model.temperature_aux(T)
        P = T_P

    nc = model.nc
    X = np.zeros((3, nc))
    X[:, index] = X0
    lnphi = np.zeros_like(X)

    global vg, Xassg

    for i, state in enumerate(equilibrium):
        out = model.logfugef_aux(X[i], temp_aux, P, state, vg[i], Xassg[i])
        lnphi[i], vg[i], Xassg[i] = out
        if np.any(np.isnan(lnphi[i])) or np.isnan(vg[i]):
            out = model.logfugef_aux(X[i], temp_aux, P, state)
            lnphi[i], vg[i], Xassg[i] = out

    lnK = lnphi[0] - lnphi[1:]
    K = np.exp(lnK)
    return np.hstack([(K[:, index]*X0[0]-X0[1:]).flatten(), X0.sum(axis=1)-1.])


def haz_objt(inc, temp_aux, P, model):

    X, W, Y = np.split(inc, 3)

    global vg, Xassg

    fugX, vg[0], Xassg[0] = model.logfugef_aux(X, temp_aux, P, 'L', vg[0],
                                               Xassg[0])
    if np.any(np.isnan(fugX)) or np.isnan(vg[0]):
        fugX, vg[0], Xassg[0] = model.logfugef_aux(X, temp_aux, P, 'L')

    fugW, vg[1], Xassg[1] = model.logfugef_aux(W, temp_aux, P, 'L', vg[1],
                                               Xassg[1])
    if np.any(np.isnan(fugW)) or np.isnan(vg[1]):
        fugW, vg[1], Xassg[1] = model.logfugef_aux(W, temp_aux, P, 'L')

    fugY, vg[2], Xassg[2] = model.logfugef_aux(Y, temp_aux, P, 'V', vg[2],
                                               Xassg[2])
    if np.any(np.isnan(fugY)) or np.isnan(vg[2]):
        fugY, vg[2], Xassg[2] = model.logfugef_aux(Y, temp_aux, P, 'V')

    K1 = np.exp(fugX-fugY)
    K2 = np.exp(fugX-fugW)

    return np.hstack([K1*X-Y, K2*X-W, X.sum()-1, Y.sum()-1, W.sum()-1])


def haz(X0, W0, Y0, T, P, model, good_initial=False, v0=[None, None, None],
        Xass0=[None, None, None], K_tol=1e-10, nacc=5, full_output=False, 
        tetha_max=10.):
    """
    Liquid liquid vapor (T,P) -> (x, w, y)

    Computes liquid liquid vapor equilibrium of multicomponent mixtures at
    given temperature. This functions uses a simultaneous method to check
    stability and equilibrium, when slow convergence is noted, minimization of
    Gibbs free energy is performed with BFGS.

    Parameters
    ----------

    X0 : array_like
         guess composition of liquid 1
    W0 : array_like
         guess composition of liquid 2
    Y0 : array_like
         guess composition of vapour 1
    T : float
        absolute temperature [K].
    P : float
        pressure [Pa]
    model : object
        created from mixture and saftvrmie function
    good_initial: bool, optional
        if True skip Gupta's method and solves full system of equations.
    v0 : list, optional
         if supplied volume used as initial value to compute fugacities
    K_tol : float, optional
            Desired accuracy of K (= X/Xr) vector
    nacc : int, optional
        number of accelerated successive substitution cycles to perform
    full_output: bool, optional
        wheter to outputs all calculation info
    tetha_max : float, optional
        maximum value for stability variables. Default is 10.

    Returns
    -------
    X : array_like
        liquid1 mole fraction vector
    W : array_like
        liquid2 mole fraction vector
    Y : array_like
        vapour mole fraction fector

    """

    nc = model.nc
    if len(X0) != nc or len(W0) != nc or len(Y0) != nc:
        raise Exception('Composition vector lenght must be equal to nc')

    Z0 = (X0+Y0+W0)/3
    nonzero = np.count_nonzero(Z0)
    x0 = np.vstack([X0, W0, Y0])
    b0 = np.array([0.33, 0.33, 0.33, 0., 0.])

    global vg, Xassg
    vg = v0.copy()
    Xassg = Xass0.copy()
    # check for binary mixture
    if nonzero == 2:
        warn('Global mixture is a binary mixture, updating temperature')
        index = np.nonzero(Z0)[0]
        sol = np.zeros_like(x0)
        inc0 = np.hstack([x0[:, index].flatten(), T])
        sol1 = root(haz_objb, inc0, args=(P, 'P', model, index, 'LLV'))
        T = sol1.x[-1]
        Xs = sol1.x[:-1].reshape(3, 2)
        sol[:, index] = Xs
        X, W, Y = sol

        if full_output:
            temp_aux = model.temperature_aux(T)
            for i, state in enumerate(['L', 'L', 'V']):
                rho, Xassg[i] = model.density_aux(sol[i], temp_aux, P, state,
                                                  rho0=1/vg[i], Xass0=Xassg[i])
                vg[i] = 1./rho

            info = {'T': T, 'P': P, 'X': sol, 'v': vg, 'Xass': Xassg,
                    'states': ['L', 'L', 'V'], 'success': sol1.success}
            out = EquilibriumResult(info)
            return out
        return X, W, Y, T

    if not good_initial:
        out = multiflash(x0, b0, ['L', 'L', 'V'], Z0, T, P, model, vg, Xassg,
                         K_tol, nacc, True)
    else:
        temp_aux = model.temperature_aux(T)
        sol = fsolve(haz_objt, x0.flatten(), args=(temp_aux, P, model))
        x0 = sol.reshape([model.nc, 3])
        Z0 = x0.mean(axis=0)
        out = multiflash(x0, b0, ['L', 'L', 'V'], Z0, T, P, model,
                         vg, Xassg, K_tol, nacc, True, tetha_max)

    Xm, beta, tetha, equilibrio = out.X, out.beta, out.tetha, out.states
    error_inner = out.error_inner
    v = out.v
    Xass = out.Xass

    if error_inner > 1e-6:
        order = [2, 0, 1]  # Y, X, W
        Xm = Xm[order]
        betatetha = np.hstack([beta[order], tetha])
        equilibrio = np.asarray(equilibrio)[order]
        v0 = np.asarray(v)[order]
        Xass0 = np.asarray(out.Xass)[order]
        out = multiflash(Xm, betatetha, equilibrio, Z0, T, P, model, v0, Xass0,
                         K_tol, nacc, full_output=True, tetha_max=tetha_max)
        order = [1, 2, 0]
        Xm, beta, tetha, equilibrio = out.X, out.beta, out.tetha, out.states
        error_inner = out.error_inner
        if error_inner > 1e-6:
            order = [2, 1, 0]  # W, X, Y
            Xm = Xm[order]
            betatetha = np.hstack([beta[order], tetha])
            equilibrio = np.asarray(equilibrio)[order]
            v0 = np.asarray(out.v)[order]
            Xass0 = np.asarray(out.Xass)[order]
            out = multiflash(Xm, betatetha, equilibrio, Z0, T, P, model, v0,
                             Xass0, K_tol, nacc, full_output=True,
                             tetha_max=tetha_max)
            order = [1, 0, 2]
            Xm, beta, tetha = out.X, out.beta, out.tetha
            equilibrio = out.states
            error_inner = out.error_inner
        Xm = Xm[order]
        beta = beta[order]
        tetha = np.hstack([0., tetha])
        tetha = tetha[order]
        v = (out.v)[order]
        Xass = np.asarray(out.Xass)[order]
    else:
        tetha = np.hstack([0., tetha])

    if full_output:
        info = {'T': T, 'P': P, 'error_outer': out.error_outer,
                'error_inner': error_inner, 'iter': out.iter,
                'beta': beta, 'tetha': tetha, 'X': Xm, 'v': v,
                'Xass': Xass, 'states': ['L', 'L', 'V'],
                'method': out.method}
        out = EquilibriumResult(info)
        return out

    tethainestable = tetha > 0.
    Xm[tethainestable] = None
    X, W, Y = Xm

    return X, W, Y


def vlle(X0, W0, Y0, Z, T, P, model, v0=[None, None, None],
         Xass0=[None, None, None], K_tol=1e-10, nacc=5,
         full_output=False, tetha_max=10.):
    """
    Liquid liquid vapor Multiflash (Z, T, P) -> (x, w, y)

    Computes liquid liquid vapor equilibrium of multicomponent mixtures at
    given temperature. This functions uses a simultaneous method to check
    stability and equilibrium, when slow convergence is noted, minimization
    of Gibbs free energy is performed with BFGS.

    Parameters
    ----------

    X0 : array_like
         guess composition of liquid 1
    W0 : array_like
         guess composition of liquid 2
    Y0 : array_like
         guess composition of vapour 1
    T : float
        absolute temperature in K.
    P : float
        pressure in bar
    model : object
        Created from mixture, eos and mixrule
    good_initial: bool, optional
        if True skip Gupta's method and solves full system of equations.
    v0 : list, optional
         if supplied volume used as initial value to compute fugacities
    K_tol : float, optional
        Desired accuracy of K (= X/Xr) vector
    nacc : int, optional
        number of accelerated successive substitution cycles to perform
    full_output: bool, optional
        wheter to outputs all calculation info
    tetha_max : float, optional
        maximum value for stability variables. Default is 10.

    Returns
    -------
    X : array_like
        liquid1 mole fraction vector
    W : array_like
        liquid2 mole fraction vector
    Y : array_like
        vapour mole fraction fector

    """

    nc = model.nc
    if len(X0) != nc or len(W0) != nc or len(Y0) != nc:
        raise Exception('Composition vector lenght must be equal to nc')

    nonzero = np.count_nonzero(Z)
    x0 = np.vstack([X0, W0, Y0])
    b0 = np.array([0.33, 0.33, 0.33, 0., 0.])

    # check for binary mixture
    if nonzero == 2:
        warn('Global mixture is a binary mixture, updating temperature')
        index = np.nonzero(Z)[0]
        sol = np.zeros_like(x0)
        global vg, Xassg
        vg = v0.copy()
        Xassg = Xass0.copy()
        inc0 = np.hstack([x0[:, index].flatten(), T])
        sol1 = root(haz_objb, inc0, args=(P, 'P', model, index, 'LLV'))
        T = sol1.x[-1]
        Xs = sol1.x[:-1].reshape(3, 2)
        sol[:, index] = Xs
        X, W, Y = sol
        if full_output:

            temp_aux = model.temperature_aux(T)
            for i, state in enumerate(['L', 'L', 'V']):
                rho, Xassg[i] = model.density_aux(sol[i], temp_aux, P, state,
                                                  rho0=1/vg[i], Xass0=Xassg[i])
                vg[i] = 1./rho

            info = {'T': T, 'P': P, 'X': sol, 'v': vg, 'Xass': Xassg,
                    'states': ['L', 'L', 'V'], 'success': sol1.success}
            out = EquilibriumResult(info)
            return out
        return X, W, Y, T
    out = multiflash(x0, b0, ['L', 'L', 'V'], Z, T, P, model, v0, Xass0,
                     K_tol, nacc, True, tetha_max)

    Xm, beta, tetha, equilibrio = out.X, out.beta, out.tetha, out.states
    error_inner = out.error_inner
    v = out.v
    Xass = out.Xass

    if error_inner > 1e-6:
        order = [2, 0, 1]  # Y, X, W
        Xm = Xm[order]
        betatetha = np.hstack([beta[order], tetha])
        equilibrio = np.asarray(equilibrio)[order]
        v0 = np.asarray(v)[order]
        Xass0 = np.asarray(out.Xass)[order]
        out = multiflash(Xm, betatetha, equilibrio, Z, T, P, model, v0,
                         Xass, K_tol, nacc, full_output=True, 
                         tetha_max=tetha_max)
        order = [1, 2, 0]
        Xm, beta, tetha, equilibrio = out.X, out.beta, out.tetha, out.states
        error_inner = out.error_inner
        if error_inner > 1e-6:
            order = [2, 1, 0]  # W, X, Y
            Xm = Xm[order]
            betatetha = np.hstack([beta[order], tetha])
            equilibrio = np.asarray(equilibrio)[order]
            v0 = np.asarray(out.v)[order]
            Xass0 = np.asarray(out.Xass)[order]
            out = multiflash(Xm, betatetha, equilibrio, Z, T, P, model, v0,
                             Xass0, K_tol, nacc, full_output=True, 
                             tetha_max=tetha_max)
            order = [1, 0, 2]
            Xm, beta, tetha = out.X, out.beta, out.tetha
            equilibrio = out.states
            error_inner = out.error_inner
        Xm = Xm[order]
        beta = beta[order]
        tetha = np.hstack([0., tetha])
        tetha = tetha[order]
        v = (out.v)[order]
        Xass = np.asarray(out.Xass)[order]
    else:
        tetha = np.hstack([0., tetha])

    if full_output:
        info = {'T': T, 'P': P, 'error_outer': out.error_outer,
                'error_inner': error_inner, 'iter': out.iter, 'beta': beta,
                'tetha': tetha, 'X': Xm, 'v': v, 'Xass': Xass,
                'states': ['L', 'L', 'V'], 'method': out.method}
        out = EquilibriumResult(info)
        return out

    tethainestable = tetha > 0.
    Xm[tethainestable] = None
    X, W, Y = Xm

    return X, W, Y


__all__ = ['haz', 'vlle']



from __future__ import division, print_function, absolute_import
import numpy as np


h = 6.626070150e-34  # J s
me = 9.10938291e-31  # 1/Kg


# Equation 68
def aideal(rho, beta):
    broglie_vol = h / np.sqrt(2*np.pi * me / beta)
    a = np.log(rho * broglie_vol**3) - 1
    return a


def daideal_drho(rho, beta):
    broglie_vol = h / np.sqrt(2*np.pi * me / beta)
    a = np.log(rho * broglie_vol**3) - 1
    da = 1./rho
    return np.hstack([a, da])


def d2aideal_drho(rho, beta):
    broglie_vol = h / np.sqrt(2*np.pi * me / beta)
    a = np.log(rho * broglie_vol**3) - 1
    da = 1./rho
    d2a = -1/rho**2
    return np.hstack([a, da, d2a])



from __future__ import division, print_function, absolute_import
import numpy as np
from scipy.optimize import root
from scipy.integrate import cumulative_trapezoid
from .cijmix_cy import cmix_cy
from .tensionresult import TensionResult
from ..constants import kb, Na

R = Na * kb


def fobj_sk(inc, spath, temp_aux, mu0, ci, sqrtci, model):
    ro = np.abs(inc[:-1])
    alpha = inc[-1]
    global Xass
    mu, Xass = model.muad_aux(ro, temp_aux, Xass)
    obj = np.zeros_like(inc)
    obj[:-1] = (mu - mu0) + alpha*sqrtci
    obj[-1] = spath - sqrtci.dot(ro)
    return obj


def ten_beta0_sk(rho1, rho2, Tsat, Psat, model, n=200, full_output=False,
                 alpha0=None, check_eq=True):

    nc = model.nc

    # Dimensionless variables
    Tfactor, Pfactor, rofactor, tenfactor, zfactor = model.sgt_adim(Tsat)
    Pad = Psat*Pfactor
    ro1a = rho1*rofactor
    ro2a = rho2*rofactor

    cij = model.ci(Tsat)
    cij /= cij[0, 0]
    ci = np.diag(cij)
    sqrtci = np.sqrt(ci)

    temp_aux = model.temperature_aux(Tsat)

    mu0, Xass01 = model.muad_aux(ro1a, temp_aux)
    mu02, Xass02 = model.muad_aux(ro2a, temp_aux)
    if check_eq:
        if not np.allclose(mu0, mu02, rtol=1e-3):
            raise Exception('Not equilibria compositions, mu1 != mu2')

    # Path function
    s0 = sqrtci.dot(ro1a)
    s1 = sqrtci.dot(ro2a)
    spath = np.linspace(s0, s1, n)
    ro = np.zeros([nc, n])
    alphas = np.zeros(n)
    ro[:, 0] = ro1a
    ro[:, -1] = ro2a

    deltaro = ro2a - ro1a
    i = 1
    r0 = ro1a + deltaro * (spath[i]-s0) / n
    # #
    # RT = R*Tsat
    #
    global Xass
    Xass = Xass01
    if alpha0 is None:
        mu, Xass = model.muad_aux(r0, temp_aux, Xass)
        alpha0 = np.mean((mu - mu0)/sqrtci)
    r0 = np.hstack([r0, alpha0])
    ro0 = root(fobj_sk, r0, args=(spath[i], temp_aux, mu0, ci, sqrtci, model),
               method='lm')
    ro0 = ro0.x
    ro[:, i] = np.abs(ro0[:-1])

    for i in range(1, n):
        sol = root(fobj_sk, ro0, args=(spath[i], temp_aux, mu0, ci, sqrtci,
                   model), method='lm')
        ro0 = sol.x
        alphas[i] = ro0[-1]
        ro[:, i] = np.abs(ro0[:-1])

    # Derivatives respect to path function
    drods = np.gradient(ro, spath, edge_order=2, axis=1)

    suma = cmix_cy(drods, cij)
    dom = np.zeros(n)
    Xass = Xass01
    for k in range(1, n - 1):
        dom[k], Xass = model.dOm_aux(ro[:, k], temp_aux, mu0, Pad, Xass)

    integral = np.nan_to_num(np.sqrt(2*dom*suma))
    tension = np.abs(np.trapz(integral, spath))
    tension *= tenfactor

    if full_output:
        # Zprofile
        with np.errstate(divide='ignore'):
            intz = (np.sqrt(suma/(2*dom)))
        intz[np.isinf(intz)] = 0
        z = np.abs(cumulative_trapezoid(intz, spath, initial=0))
        z /= zfactor
        ro /= rofactor
        dictresult = {'tension': tension, 'rho': ro, 'z': z,
                      'GPT': dom, 'path': spath, 'alphas': alphas}
        out = TensionResult(dictresult)
        return out

    return tension



from __future__ import division, print_function, absolute_import
import numpy as np
from .monomer_aux import xhi_eff, dxhieff_dxhi00, d2xhieff_dxhi00
from .monomer_aux import d3xhieff_dxhi00, dxhieff_dx, dxhieff_dx_dxhi00
from .monomer_aux import dxhieff_dx_dxhi00_dxxhi, dxhieff_dx_d2xhi00_dxxhi


# Eq (19) Paper 2014
def a1s(xhi00, xhix_vec, xm, cictes, a1vdw):
    xhieff = xhi_eff(xhix_vec, cictes)
    ghs = (1 - xhieff/2) / (1 - xhieff)**3
    a1 = a1vdw * ghs * xhi00 * xm
    return a1


def da1s_dxhi00(xhi00, xhix_vec, xm, cictes, a1vdw, dxhix_dxhi00):
    # xhieff = xhi_eff(xhix, cictes)
    # dxhieff = dxhieff_dxhi00(xhix, cictes, dxhix_dxhi00)
    xhieff, dxhieff = dxhieff_dxhi00(xhix_vec, cictes, dxhix_dxhi00)

    xhieff_1 = 1 - xhieff
    xhieff_13 = xhieff_1**3
    xhieff_14 = xhieff_13*xhieff_1

    xhieff2 = xhieff**2
    aux1 = a1vdw * xhi00
    aux2 = (5 - 2 * xhieff)

    ghs = (1 - xhieff/2) / xhieff_13
    a1 = aux1 * ghs * xm

    da1 = 2 - 3*xhieff + xhieff2
    da1 += xhi00 * dxhieff * aux2
    da1 *= a1vdw * xm
    da1 /= 2*xhieff_14

    return a1, da1


def d2a1s_dxhi00(xhi00, xhix_vec, xm, cictes, a1vdw, dxhix_dxhi00):
    # xhieff = xhi_eff(xhix, cictes)
    # dxhieff = dxhieff_dxhi00(xhix, cictes, dxhix_dxhi00)
    # d2xhieff = d2xhieff_dxhi00(xhix, cictes, dxhix_dxhi00)
    xhieff, dxhieff, d2xhieff = d2xhieff_dxhi00(xhix_vec, cictes, dxhix_dxhi00)

    xhieff_1 = 1 - xhieff
    xhieff_13 = xhieff_1**3
    xhieff_14 = xhieff_13*xhieff_1
    xhieff_15 = xhieff_14*xhieff_1

    xhieff2 = xhieff**2
    aux1 = a1vdw * xhi00
    aux2 = (5 - 2 * xhieff)
    aux3 = 2 - 3*xhieff + xhieff2
    aux3 += xhi00 * aux2 * dxhieff
    aux3 *= -xhieff_1
    aux4 = -5 + 7*xhieff - 2*xhieff2
    aux5 = -3 + xhieff

    ghs = (1 - xhieff/2) / xhieff_13
    a1 = aux1 * ghs * xm

    da1 = 2 - 3*xhieff + xhieff2
    da1 += xhi00 * dxhieff * aux2
    da1 *= a1vdw * xm
    da1 /= 2*xhieff_14

    d2a1 = -2 * xhieff_1 * aux2 * dxhieff
    d2a1 += 6 * xhi00 * aux5 * dxhieff**2
    d2a1 += xhi00 * aux4 * d2xhieff
    d2a1 *= -a1vdw * xm
    d2a1 /= 2*xhieff_15

    return a1, da1, d2a1


def d3a1s_dxhi00(xhi00, xhix_vec, xm, cictes, a1vdw, dxhix_dxhi00):

    # xhieff = xhi_eff(xhix, cictes)
    # dxhieff = dxhieff_dxhi00(xhix, cictes, dxhix_dxhi00)
    # d2xhieff = d2xhieff_dxhi00(xhix, cictes, dxhix_dxhi00)
    # d3xhieff = d3xhieff_dxhi00(xhix, cictes, dxhix_dxhi00)
    out = d3xhieff_dxhi00(xhix_vec, cictes, dxhix_dxhi00)
    xhieff, dxhieff, d2xhieff, d3xhieff = out

    xhieff_1 = 1 - xhieff
    xhieff_13 = xhieff_1**3
    xhieff_14 = xhieff_13*xhieff_1
    xhieff_15 = xhieff_14*xhieff_1
    xhieff_16 = xhieff_15*xhieff_1

    xhieff2 = xhieff**2
    aux1 = a1vdw * xhi00
    aux2 = (5 - 2 * xhieff)
    aux3 = 2 - 3*xhieff + xhieff2
    aux3 += xhi00 * aux2 * dxhieff
    aux3 *= -xhieff_1
    aux4 = -5 + 7*xhieff - 2*xhieff2
    aux5 = -3 + xhieff

    ghs = (1 - xhieff/2) / xhieff_13
    a1 = aux1 * ghs * xm

    da1 = 2 - 3*xhieff + xhieff2
    da1 += xhi00 * dxhieff * aux2
    da1 *= a1vdw * xm
    da1 /= 2*xhieff_14

    d2a1 = -2 * xhieff_1 * aux2 * dxhieff
    d2a1 += 6 * xhi00 * aux5 * dxhieff**2
    d2a1 += xhi00 * aux4 * d2xhieff
    d2a1 *= -a1vdw * xm
    d2a1 /= 2*xhieff_15

    d3a1 = 18 * aux5 * -xhieff_1 * dxhieff**2
    d3a1 += 12 * xhi00 * (7 - 2 * xhieff) * dxhieff**3
    d3a1 += 18 * xhi00 * aux5 * -xhieff_1 * dxhieff * d2xhieff
    d3a1 += xhieff_1**2 * aux2 * (3*d2xhieff+xhi00*d3xhieff)
    d3a1 *= a1vdw * xm
    d3a1 /= 2*xhieff_16

    return a1, da1, d2a1, d3a1


def da1s_dx(xhi00, xhix_vec, xm, ms, cictes, a1vdw, dxhix_dx):

    # xhieff = xhi_eff(xhix, cictes)
    # dxhieff = dxhieff_dx(xhix, cictes, dxhix_dx)
    xhieff, dxhieff = dxhieff_dx(xhix_vec, cictes, dxhix_dx)
    xhieff_1 = 1 - xhieff
    xhieff_13 = xhieff_1**3
    xhieff_14 = xhieff_13*xhieff_1

    aux1 = a1vdw * xhi00

    ghs = (1 - xhieff/2) / xhieff_13
    a1 = aux1 * ghs * xm

    da1 = np.multiply.outer(ms, (-2 + xhieff) * -xhieff_1)
    da1 += (5 - 2 * xhieff) * xm * dxhieff
    da1 *= aux1
    da1 /= 2*xhieff_14
    return a1, da1


def da1s_dx_dxhi00(xhi00, xhix_vec, xm, ms, cictes, a1vdw, dxhix_dxhi00,
                   dxhix_dx):

    # xhieff = xhi_eff(xhix, cictes)
    # dxhieff = dxhieff_dxhi00(xhix, cictes, dxhix_dxhi00)
    # dxhieffdx = dxhieff_dx(xhix, cictes, dxhix_dx)

    out = dxhieff_dx_dxhi00(xhix_vec, cictes, dxhix_dxhi00, dxhix_dx)
    xhieff, dxhieff, dxhieff_dx = out
    xhieff_1 = 1 - xhieff
    xhieff_13 = xhieff_1**3
    xhieff_14 = xhieff_13*xhieff_1

    aux1 = a1vdw * xhi00
    aux2 = (5 - 2 * xhieff)

    ghs = (1 - xhieff/2) / xhieff_13
    a1 = aux1 * ghs * xm

    da1 = 2 - 3*xhieff + xhieff**2
    da1 += xhi00 * dxhieff * aux2
    da1 *= a1vdw * xm
    da1 /= 2*xhieff_14

    da1x = np.multiply.outer(ms, (2 - xhieff) * xhieff_1)
    da1x += aux2 * xm * dxhieff_dx
    da1x *= aux1
    da1x /= 2*xhieff_14
    return a1, da1, da1x


def da1s_dx_dxhi00_dxxhi(xhi00, xhix_vec, xm, ms, cictes, a1vdw,
                         dxhix_dxhi00, dxhix_dx,  dxhix_dx_dxhi00):

    # xhieff = xhi_eff(xhix, cictes)
    # dxhieffdx = dxhieff_dx(xhix, cictes, dxhix_dx)
    # dxhieff = dxhieff_dxhi00(xhix, cictes, dxhix_dxhi00)
    # dxhieff_dxdxhi00 = dxhieff_dx_dxhi00(xhix, cictes, dxhix_dxhi00,
    #                                     dxhix_dx, dxhix_dx_dxhi00)

    out = dxhieff_dx_dxhi00_dxxhi(xhix_vec, cictes, dxhix_dxhi00, dxhix_dx,
                                  dxhix_dx_dxhi00)
    xhieff, dxhieff, dxhieff_dx, dxhieff_dxxhi = out

    xhieff_1 = 1 - xhieff
    xhieff_13 = xhieff_1**3
    xhieff_14 = xhieff_13*xhieff_1
    xhieff_15 = xhieff_14*xhieff_1

    xhieff2 = xhieff**2
    aux1 = a1vdw * xhi00
    aux2 = (5 - 2 * xhieff)

    aux3 = 2 - 3*xhieff + xhieff2
    aux3 += xhi00 * aux2 * dxhieff
    aux3 *= -xhieff_1

    aux4 = -5 + 7*xhieff - 2*xhieff2

    ghs = (1 - xhieff/2) / xhieff_13
    a1 = aux1 * ghs * xm

    da1 = 2 - 3*xhieff + xhieff2
    da1 += xhi00 * dxhieff * aux2
    da1 *= a1vdw * xm
    da1 /= 2*xhieff_14

    da1x = np.multiply.outer(ms, (2 - xhieff) * xhieff_1)
    da1x += aux2 * xm * dxhieff_dx
    da1x *= aux1
    da1x /= 2*xhieff_14

    da1xxhi = dxhieff_dx
    da1xxhi *= aux4 + 6*xhi00 * (-3 + xhieff)*dxhieff
    da1xxhi += xhi00 * dxhieff_dxxhi * aux4
    da1xxhi *= xm
    da1xxhi += np.multiply.outer(ms, aux3)
    da1xxhi *= a1vdw
    da1xxhi /= -2*xhieff_15
    return a1, da1, da1x, da1xxhi


def da1s_dx_d2xhi00_dxxhi(xhi00, xhix_vec, xm, ms, cictes, a1vdw,
                          dxhix_dxhi00, dxhix_dx,  dxhix_dx_dxhi00):

    # xhieff = xhi_eff(xhix, cictes)
    # dxhieff = dxhieff_dxhi00(xhix, cictes, dxhix_dxhi00)
    # d2xhieff = d2xhieff_dxhi00(xhix, cictes, dxhix_dxhi00)
    # dxhieffdx = dxhieff_dx(xhix, cictes, dxhix_dx)
    # dxhieff_dxdxhi00 = dxhieff_dx_dxhi00(xhix, cictes, dxhix_dxhi00,
    #                                      dxhix_dx, dxhix_dx_dxhi00)

    out = dxhieff_dx_d2xhi00_dxxhi(xhix_vec, cictes, dxhix_dxhi00, dxhix_dx,
                                   dxhix_dx_dxhi00)
    xhieff, dxhieff, d2xhieff, dxhieff_dx, dxhieff_dxxhi = out

    xhieff_1 = 1 - xhieff
    xhieff_13 = xhieff_1**3
    xhieff_14 = xhieff_13*xhieff_1
    xhieff_15 = xhieff_14*xhieff_1

    xhieff2 = xhieff**2
    aux1 = a1vdw * xhi00
    aux2 = (5 - 2 * xhieff)
    aux3 = 2 - 3*xhieff + xhieff2
    aux3 += xhi00 * aux2 * dxhieff
    aux3 *= -xhieff_1
    aux4 = -5 + 7*xhieff - 2*xhieff2
    aux5 = -3 + xhieff

    ghs = (1 - xhieff/2) / xhieff_13
    a1 = aux1 * ghs * xm

    da1 = 2 - 3*xhieff + xhieff2
    da1 += xhi00 * dxhieff * aux2
    da1 *= a1vdw * xm
    da1 /= 2*xhieff_14

    d2a1 = -2 * xhieff_1 * aux2 * dxhieff
    d2a1 += 6 * xhi00 * aux5 * dxhieff**2
    d2a1 += xhi00 * aux4 * d2xhieff
    d2a1 *= -a1vdw * xm
    d2a1 /= 2*xhieff_15

    da1x = np.multiply.outer(ms, (2 - xhieff) * xhieff_1)
    da1x += aux2 * xm * dxhieff_dx
    da1x *= aux1
    da1x /= 2*xhieff_14

    da1xxhi = dxhieff_dx
    da1xxhi *= aux4 + 6*xhi00 * aux5 * dxhieff
    da1xxhi += xhi00 * dxhieff_dxxhi * aux4
    da1xxhi *= xm
    da1xxhi += np.multiply.outer(ms, aux3)
    da1xxhi *= a1vdw
    da1xxhi /= -2*xhieff_15

    return a1, da1, d2a1, da1x, da1xxhi



from __future__ import division, print_function, absolute_import
import numpy as np
from .B_monomer import B, dB_dxhi00, d2B_dxhi00, d3B_dxhi00
from .a1s_monomer import a1s, da1s_dxhi00, d2a1s_dxhi00, d3a1s_dxhi00


def a1sB(xhi00, xhix, xhix_vec, xm, Ikl, Jkl, cictes, a1vdw, a1vdw_cte):
    a1 = a1s(xhi00, xhix_vec, xm, cictes, a1vdw)
    b = B(xhi00, xhix, xm, Ikl, Jkl, a1vdw_cte)
    return a1 + b


def da1sB_dxhi00(xhi00, xhix, xhix_vec, xm, Ikl, Jkl, cictes, a1vdw, a1vdw_cte,
                 dxhix_dxhi00):
    a1, da1 = da1s_dxhi00(xhi00, xhix_vec, xm, cictes, a1vdw, dxhix_dxhi00)
    b, db = dB_dxhi00(xhi00, xhix, xm, Ikl, Jkl, a1vdw_cte, dxhix_dxhi00)
    return a1 + b, da1 + db


def d2a1sB_dxhi00(xhi00, xhix, xhix_vec, xm, Ikl, Jkl, cictes, a1vdw,
                  a1vdw_cte, dxhix_dxhi00):
    a1, da1, d2a1 = d2a1s_dxhi00(xhi00, xhix_vec, xm, cictes, a1vdw,
                                 dxhix_dxhi00)
    b, db, d2b = d2B_dxhi00(xhi00, xhix, xm, Ikl, Jkl, a1vdw_cte, dxhix_dxhi00)
    return a1 + b, da1 + db, d2a1 + d2b


def d3a1sB_dxhi00(xhi00, xhix, xhix_vec, xm, Ikl, Jkl, cictes, a1vdw,
                  a1vdw_cte, dxhix_dxhi00):
    a1, da1, d2a1, d3a1 = d3a1s_dxhi00(xhi00, xhix_vec, xm, cictes, a1vdw,
                                       dxhix_dxhi00)
    b, db, d2b, d3b = d3B_dxhi00(xhi00, xhix, xm, Ikl, Jkl, a1vdw_cte,
                                 dxhix_dxhi00)
    return a1 + b, da1 + db, d2a1 + d2b, d3a1 + d3b


def a1sB_eval(xhi00, xhix, xhix_vec, xs_m, I_lambdaskl, J_lambdaskl,
              ccteskl, a1vdwkl, a1vdw_ctekl):

    # la_kl, lr_kl, la_kl2, lr_kl2, lar_kl = lambdaskl
    cctes_lakl, cctes_lrkl, cctes_2lakl, cctes_2lrkl, cctes_larkl = ccteskl
    a1vdw_lakl, a1vdw_lrkl, a1vdw_2lakl, a1vdw_2lrkl, a1vdw_larkl = a1vdwkl
    I_lakl, I_lrkl, I_2lakl, I_2lrkl, I_larkl = I_lambdaskl
    J_lakl, J_lrkl, J_2lakl, J_2lrkl, J_larkl = J_lambdaskl

    a1sb_a = a1sB(xhi00, xhix, xhix_vec, xs_m, I_lakl, J_lakl, cctes_lakl,
                  a1vdw_lakl, a1vdw_ctekl)
    a1sb_r = a1sB(xhi00, xhix, xhix_vec, xs_m, I_lrkl, J_lrkl, cctes_lrkl,
                  a1vdw_lrkl, a1vdw_ctekl)
    a1sb_2a = a1sB(xhi00, xhix, xhix_vec, xs_m, I_2lakl, J_2lakl, cctes_2lakl,
                   a1vdw_2lakl, a1vdw_ctekl)
    a1sb_2r = a1sB(xhi00, xhix, xhix_vec, xs_m, I_2lrkl, J_2lrkl, cctes_2lrkl,
                   a1vdw_2lrkl, a1vdw_ctekl)
    a1sb_ar = a1sB(xhi00, xhix, xhix_vec, xs_m, I_larkl, J_larkl, cctes_larkl,
                   a1vdw_larkl, a1vdw_ctekl)

    a1sb_a1 = np.array([a1sb_a, a1sb_r])
    a1sb_a2 = np.array([a1sb_2a, a1sb_ar, a1sb_2r])
    return a1sb_a1, a1sb_a2


def da1sB_dxhi00_eval(xhi00, xhix, xhix_vec, xs_m, I_lambdaskl, J_lambdaskl,
                      ccteskl, a1vdwkl, a1vdw_ctekl, dxhix_dxhi00):

    # la_kl, lr_kl, la_kl2, lr_kl2, lar_kl = lambdaskl
    cctes_lakl, cctes_lrkl, cctes_2lakl, cctes_2lrkl, cctes_larkl = ccteskl
    a1vdw_lakl, a1vdw_lrkl, a1vdw_2lakl, a1vdw_2lrkl, a1vdw_larkl = a1vdwkl
    I_lakl, I_lrkl, I_2lakl, I_2lrkl, I_larkl = I_lambdaskl
    J_lakl, J_lrkl, J_2lakl, J_2lrkl, J_larkl = J_lambdaskl

    a1sb_a, da1sb_a = da1sB_dxhi00(xhi00, xhix, xhix_vec, xs_m, I_lakl, J_lakl,
                                   cctes_lakl, a1vdw_lakl, a1vdw_ctekl,
                                   dxhix_dxhi00)

    a1sb_r, da1sb_r = da1sB_dxhi00(xhi00, xhix, xhix_vec, xs_m, I_lrkl, J_lrkl,
                                   cctes_lrkl, a1vdw_lrkl, a1vdw_ctekl,
                                   dxhix_dxhi00)

    a1sb_2a, da1sb_2a = da1sB_dxhi00(xhi00, xhix, xhix_vec, xs_m, I_2lakl,
                                     J_2lakl, cctes_2lakl, a1vdw_2lakl,
                                     a1vdw_ctekl, dxhix_dxhi00)

    a1sb_2r, da1sb_2r = da1sB_dxhi00(xhi00, xhix, xhix_vec, xs_m, I_2lrkl,
                                     J_2lrkl, cctes_2lrkl, a1vdw_2lrkl,
                                     a1vdw_ctekl, dxhix_dxhi00)

    a1sb_ar, da1sb_ar = da1sB_dxhi00(xhi00, xhix, xhix_vec, xs_m, I_larkl,
                                     J_larkl, cctes_larkl, a1vdw_larkl,
                                     a1vdw_ctekl, dxhix_dxhi00)

    a1sb_a1 = np.array([[a1sb_a, a1sb_r],
                        [da1sb_a, da1sb_r]])
    a1sb_a2 = np.array([[a1sb_2a, a1sb_ar, a1sb_2r],
                       [da1sb_2a, da1sb_ar, da1sb_2r]])

    return a1sb_a1, a1sb_a2


def d2a1sB_dxhi00_eval(xhi00, xhix, xhix_vec, xs_m, I_lambdaskl, J_lambdaskl,
                       ccteskl, a1vdwkl, a1vdw_ctekl, dxhix_dxhi00):

    # la_kl, lr_kl, la_kl2, lr_kl2, lar_kl = lambdaskl
    cctes_lakl, cctes_lrkl, cctes_2lakl, cctes_2lrkl, cctes_larkl = ccteskl
    a1vdw_lakl, a1vdw_lrkl, a1vdw_2lakl, a1vdw_2lrkl, a1vdw_larkl = a1vdwkl

    I_lakl, I_lrkl, I_2lakl, I_2lrkl, I_larkl = I_lambdaskl
    J_lakl, J_lrkl, J_2lakl, J_2lrkl, J_larkl = J_lambdaskl

    out_la = d2a1sB_dxhi00(xhi00, xhix, xhix_vec, xs_m, I_lakl, J_lakl,
                           cctes_lakl, a1vdw_lakl, a1vdw_ctekl, dxhix_dxhi00)
    a1sb_a, da1sb_a, d2a1sb_a = out_la

    out_lr = d2a1sB_dxhi00(xhi00, xhix, xhix_vec, xs_m, I_lrkl, J_lrkl,
                           cctes_lrkl, a1vdw_lrkl, a1vdw_ctekl, dxhix_dxhi00)
    a1sb_r, da1sb_r, d2a1sb_r = out_lr

    out_2la = d2a1sB_dxhi00(xhi00, xhix, xhix_vec, xs_m, I_2lakl, J_2lakl,
                            cctes_2lakl, a1vdw_2lakl, a1vdw_ctekl,
                            dxhix_dxhi00)
    a1sb_2a, da1sb_2a, d2a1sb_2a = out_2la

    out_2lr = d2a1sB_dxhi00(xhi00, xhix, xhix_vec, xs_m, I_2lrkl, J_2lrkl,
                            cctes_2lrkl, a1vdw_2lrkl, a1vdw_ctekl,
                            dxhix_dxhi00)
    a1sb_2r, da1sb_2r, d2a1sb_2r = out_2lr

    out_lar = d2a1sB_dxhi00(xhi00, xhix, xhix_vec, xs_m, I_larkl, J_larkl,
                            cctes_larkl, a1vdw_larkl, a1vdw_ctekl,
                            dxhix_dxhi00)
    a1sb_ar, da1sb_ar, d2a1sb_ar = out_lar

    a1sb_a1 = np.array([[a1sb_a, a1sb_r],
                        [da1sb_a, da1sb_r],
                       [d2a1sb_a, d2a1sb_r]])

    a1sb_a2 = np.array([[a1sb_2a, a1sb_ar, a1sb_2r],
                       [da1sb_2a, da1sb_ar, da1sb_2r],
                       [d2a1sb_2a, d2a1sb_ar, d2a1sb_2r]])
    return a1sb_a1, a1sb_a2


def d3a1sB_dxhi00_eval(xhi00, xhix, xhix_vec, xs_m, I_lambdaskl, J_lambdaskl,
                       ccteskl, a1vdwkl, a1vdw_ctekl, dxhix_dxhi00):

    cctes_lakl, cctes_lrkl, cctes_2lakl, cctes_2lrkl, cctes_larkl = ccteskl
    a1vdw_lakl, a1vdw_lrkl, a1vdw_2lakl, a1vdw_2lrkl, a1vdw_larkl = a1vdwkl

    I_lakl, I_lrkl, I_2lakl, I_2lrkl, I_larkl = I_lambdaskl
    J_lakl, J_lrkl, J_2lakl, J_2lrkl, J_larkl = J_lambdaskl

    out_la = d3a1sB_dxhi00(xhi00, xhix, xhix_vec, xs_m, I_lakl, J_lakl,
                           cctes_lakl, a1vdw_lakl, a1vdw_ctekl, dxhix_dxhi00)
    a1sb_a, da1sb_a, d2a1sb_a, d3a1sb_a = out_la

    out_lr = d3a1sB_dxhi00(xhi00, xhix, xhix_vec, xs_m, I_lrkl, J_lrkl,
                           cctes_lrkl, a1vdw_lrkl, a1vdw_ctekl, dxhix_dxhi00)
    a1sb_r, da1sb_r, d2a1sb_r, d3a1sb_r = out_lr

    out_2la = d3a1sB_dxhi00(xhi00, xhix, xhix_vec, xs_m, I_2lakl, J_2lakl,
                            cctes_2lakl, a1vdw_2lakl, a1vdw_ctekl,
                            dxhix_dxhi00)
    a1sb_2a, da1sb_2a, d2a1sb_2a, d3a1sb_2a = out_2la

    out_2lr = d3a1sB_dxhi00(xhi00, xhix, xhix_vec, xs_m, I_2lrkl, J_2lrkl,
                            cctes_2lrkl, a1vdw_2lrkl, a1vdw_ctekl,
                            dxhix_dxhi00)
    a1sb_2r, da1sb_2r, d2a1sb_2r, d3a1sb_2r = out_2lr

    out_lar = d3a1sB_dxhi00(xhi00, xhix, xhix_vec, xs_m, I_larkl, J_larkl,
                            cctes_larkl, a1vdw_larkl, a1vdw_ctekl,
                            dxhix_dxhi00)
    a1sb_ar, da1sb_ar, d2a1sb_ar, d3a1sb_ar = out_lar

    a1sb_a1 = np.array([[a1sb_a, a1sb_r],
                        [da1sb_a, da1sb_r],
                       [d2a1sb_a, d2a1sb_r],
                       [d3a1sb_a, d3a1sb_r]])

    a1sb_a2 = np.array([[a1sb_2a, a1sb_ar, a1sb_2r],
                       [da1sb_2a, da1sb_ar, da1sb_2r],
                       [d2a1sb_2a, d2a1sb_ar, d2a1sb_2r],
                       [d3a1sb_2a, d3a1sb_ar, d3a1sb_2r]])

    return a1sb_a1, a1sb_a2


def x0lambda_evalm(x0, laij, lrij, larij):
    x0la = x0**laij
    x0lr = x0**lrij
    x02la = x0**(2*laij)
    x02lr = x0**(2*lrij)
    x0lar = x0**larij

    # To be used for a1 and a2 in monomer term
    x0_a1 = np.array([x0la, -x0lr])
    x0_a2 = np.array([x02la, -2*x0lar, x02lr])

    return x0_a1, x0_a2


def x0lambda_evalc(x0, la, lr, lar):
    x0la = x0**la
    x0lr = x0**lr
    x02la = x0**(2*la)
    x02lr = x0**(2*lr)
    x0lar = x0**lar

    # To be used for a1 and a2 in monomer term
    x0_a1 = np.array([x0la, -x0lr])
    x0_a2 = np.array([x02la, -2*x0lar, x02lr])
    # To be used in g1 and g2 of chain
    x0_g1 = np.array([la * x0la, -lr*x0lr])
    x0_g2 = np.array([la * x02la, -lar*x0lar, lr * x02lr])

    return x0_a1, x0_a2, x0_g1, x0_g2



from __future__ import division, print_function, absolute_import
import numpy as np


def asso_aux(Nst_kk, sites_kk, groups_index, subgroups, df_asso_kl):

    index_asso = np.where(Nst_kk > 0)
    molecule_id_index_asso = groups_index[index_asso]
    subgroup_id_asso = subgroups[index_asso]
    sites_asso = sites_kk[index_asso]
    n_sites_molecule = Nst_kk[index_asso]

    molecule_id_index_sites = []
    for i, j in enumerate(sites_asso):
        molecule_id_index_sites += [molecule_id_index_asso[i]] * np.count_nonzero(j)
    molecule_id_index_sites = np.asarray(molecule_id_index_sites)

    group_asso_index = []
    for i, j in enumerate(sites_kk):
        group_asso_index += np.count_nonzero(j) * [i]
    group_asso_index = np.asarray(group_asso_index)

    nsites = len(group_asso_index)
    epsAB_kl = np.zeros([nsites, nsites])
    kAB_kl = np.zeros([nsites, nsites])

    sites_cumsum = np.cumsum(np.hstack([0., n_sites_molecule[:-1]]), dtype=np.int64)

    ngroups = sites_asso.shape[0]
    move_pos_asso = np.array(ngroups*[np.arange(0, 3)])
    where_0e1 = sites_asso[:, 0] == 0
    move_pos_asso[where_0e1] = np.array([0, 0, 1])
    where_0e2 = sites_asso[:, 1] == 0
    move_pos_asso[where_0e2] = np.array([0, 0, 0])

    # index of associating molecule number
    indexABij1 = []
    indexABij2 = []
    # index of interacting site numbers
    indexAB_id1 = []
    indexAB_id2 = []
    for k, groupK in enumerate(subgroup_id_asso):
        for l in range(k, ngroups):
            groupL = subgroup_id_asso[l]

            bool_kk = df_asso_kl.group_k == groupK
            bool_ll = df_asso_kl.group_l == groupL

            bool_kl = df_asso_kl.group_k == groupL

            bool_lk = df_asso_kl.group_l == groupK

            df1 = df_asso_kl[bool_kk & bool_ll]
            len1 = df1.shape[0]

            df2 = df_asso_kl[bool_kl & bool_lk]
            len2 = df2.shape[0]

            if len1 >= len2:
                df = df1
                n2 = len1
            elif len2 > len1:
                df = df2
                n2 = len2

            for j in range(n2):
                values = df.iloc[j].values[0:6]
                groupK2, siteK, groupL2, siteL, epsAB, kAB = values

                if groupK2 != groupK:
                    siteK, siteL = siteL, siteK
                    groupK2, groupL2 = groupL2, groupK2

                if siteK == 'H':
                    moveK = move_pos_asso[k, 0]
                elif siteK == 'e1':
                    moveK = move_pos_asso[k, 1]
                elif siteK == 'e2':
                    moveK = move_pos_asso[k, 2]

                if siteL == 'H':
                    moveL = move_pos_asso[l, 0]
                elif siteL == 'e1':
                    moveL = move_pos_asso[l, 1]
                elif siteL == 'e2':
                    moveL = move_pos_asso[l, 2]

                index0 = sites_cumsum[k] + moveK
                indexf = sites_cumsum[l] + moveL
                epsAB_kl[index0, indexf] = epsAB
                epsAB_kl[indexf, index0] = epsAB
                kAB_kl[index0, indexf] = kAB
                kAB_kl[indexf, index0] = kAB

                # the lenght of the list is the number of interactions
                # they should have the same length as the number of interacting sites
                indexAB_id1.append(molecule_id_index_sites[index0])
                indexAB_id2.append(molecule_id_index_sites[indexf])
                indexABij1.append(index0)
                indexABij2.append(indexf)

                molecule_id0 = molecule_id_index_sites[index0]
                molecule_idf = molecule_id_index_sites[indexf]
                self_site_associating = molecule_id0 == molecule_idf

                if siteK != siteL:
                    indexAB_id1.append(molecule_id_index_sites[indexf])
                    indexAB_id2.append(molecule_id_index_sites[index0])
                    indexABij1.append(indexf)
                    indexABij2.append(index0)
                elif siteK == siteL and not self_site_associating:
                    indexAB_id1.append(molecule_id_index_sites[indexf])
                    indexAB_id2.append(molecule_id_index_sites[index0])
                    indexABij1.append(indexf)
                    indexABij2.append(index0)

            if len1 == 0 and len2 == 0:
                # mixing rule if both groups self-associate
                dfkk = df_asso_kl[bool_kk & bool_lk]
                lenkk = dfkk.shape[0]
                dfll = df_asso_kl[bool_kl & bool_ll]
                lenll = dfll.shape[0]
                epsAB = 0.
                kAB = 0.

                if lenkk == 1 and lenll == 1:
                    values = dfkk.iloc[0].values[0:6]
                    _, siteK_kk, _, siteL_kk, epsAB_kk, kAB_kk = values

                    values = dfll.iloc[0].values[0:6]
                    _, siteK_ll, _, siteL_ll, epsAB_ll, kAB_ll = values

                    epsAB = np.sqrt(epsAB_kk * epsAB_ll)
                    kAB = ((np.cbrt(kAB_kk) + np.cbrt(kAB_ll))/2)**3

                    bool_k1 = [siteK_kk, siteL_kk] == ['H', 'e1']
                    bool_k2 = [siteL_kk, siteK_kk] == ['H', 'e1']
                    bool_k3 = [siteL_kk, siteK_kk] == ['H', 'H']
                    bool_k4 = sites_asso[k, 1] != 0
                    bool_k5 = sites_asso[k, 2] != 0

                    bool_l1 = [siteK_ll, siteL_ll] == ['H', 'e1']
                    bool_l2 = [siteL_ll, siteK_ll] == ['H', 'e1']
                    bool_l3 = [siteL_ll, siteK_ll] == ['H', 'H']
                    bool_l4 = sites_asso[l, 1] != 0
                    bool_l5 = sites_asso[l, 2] != 0

                    bool_k = bool_k1 or bool_k2
                    bool_l = bool_l1 or bool_l2

                    bool_aux1 = bool_l and bool_k3
                    bool_aux2 = bool_k and bool_l3

                    if bool_k and bool_l:
                        # siteK = 'H' and siteL = 'e1' or  siteK = 'e1' and siteL = 'H'
                        index0 = sites_cumsum[k] + move_pos_asso[k, 0]
                        indexf = sites_cumsum[l] + move_pos_asso[l, 1]
                        epsAB_kl[index0, indexf] = epsAB
                        epsAB_kl[indexf, index0] = epsAB
                        kAB_kl[index0, indexf] = kAB
                        kAB_kl[indexf, index0] = kAB

                        indexAB_id1.append(molecule_id_index_sites[index0])
                        indexAB_id2.append(molecule_id_index_sites[indexf])
                        indexABij1.append(index0)
                        indexABij2.append(indexf)
                        indexAB_id1.append(molecule_id_index_sites[indexf])
                        indexAB_id2.append(molecule_id_index_sites[index0])
                        indexABij1.append(indexf)
                        indexABij2.append(index0)
                    if bool_aux1:
                        if bool_l4:
                            # siteK = 'H' and siteL = 'e1'
                            index0 = sites_cumsum[k] + move_pos_asso[k, 0]
                            indexf = sites_cumsum[l] + move_pos_asso[l, 1]
                            epsAB_kl[index0, indexf] = epsAB
                            epsAB_kl[indexf, index0] = epsAB
                            kAB_kl[index0, indexf] = kAB
                            kAB_kl[indexf, index0] = kAB

                            indexAB_id1.append(molecule_id_index_sites[index0])
                            indexAB_id2.append(molecule_id_index_sites[indexf])
                            indexABij1.append(index0)
                            indexABij2.append(indexf)
                            indexAB_id1.append(molecule_id_index_sites[indexf])
                            indexAB_id2.append(molecule_id_index_sites[index0])
                            indexABij1.append(indexf)
                            indexABij2.append(index0)

                        if bool_l5:
                            # siteK = 'H' and siteL = 'e2'
                            index0 = sites_cumsum[k] + move_pos_asso[k, 0]
                            indexf = sites_cumsum[l] + move_pos_asso[l, 2]
                            epsAB_kl[index0, indexf] = epsAB
                            epsAB_kl[indexf, index0] = epsAB
                            kAB_kl[index0, indexf] = kAB
                            kAB_kl[indexf, index0] = kAB

                            indexAB_id1.append(molecule_id_index_sites[index0])
                            indexAB_id2.append(molecule_id_index_sites[indexf])
                            indexABij1.append(index0)
                            indexABij2.append(indexf)
                            indexAB_id1.append(molecule_id_index_sites[indexf])
                            indexAB_id2.append(molecule_id_index_sites[index0])
                            indexABij1.append(indexf)
                            indexABij2.append(index0)

                    if bool_aux2:
                        if bool_k4:
                            # siteK = 'e1' and siteL = 'H'
                            index0 = sites_cumsum[k] + move_pos_asso[k, 1]
                            indexf = sites_cumsum[l] + move_pos_asso[l, 0]
                            epsAB_kl[index0, indexf] = epsAB
                            epsAB_kl[indexf, index0] = epsAB
                            kAB_kl[index0, indexf] = kAB
                            kAB_kl[indexf, index0] = kAB

                            indexAB_id1.append(molecule_id_index_sites[index0])
                            indexAB_id2.append(molecule_id_index_sites[indexf])
                            indexABij1.append(index0)
                            indexABij2.append(indexf)
                            indexAB_id1.append(molecule_id_index_sites[indexf])
                            indexAB_id2.append(molecule_id_index_sites[index0])
                            indexABij1.append(indexf)
                            indexABij2.append(index0)

                        if bool_k5:
                            # siteK = 'e2' and siteL = 'H'
                            index0 = sites_cumsum[k] + move_pos_asso[k, 2]
                            indexf = sites_cumsum[l] + move_pos_asso[l, 0]
                            epsAB_kl[index0, indexf] = epsAB
                            epsAB_kl[indexf, index0] = epsAB
                            kAB_kl[index0, indexf] = kAB
                            kAB_kl[indexf, index0] = kAB

                            indexAB_id1.append(molecule_id_index_sites[index0])
                            indexAB_id2.append(molecule_id_index_sites[indexf])
                            indexABij1.append(index0)
                            indexABij2.append(indexf)
                            indexAB_id1.append(molecule_id_index_sites[indexf])
                            indexAB_id2.append(molecule_id_index_sites[index0])
                            indexABij1.append(indexf)
                            indexABij2.append(index0)

                    if bool_k3 and bool_l3:
                        # siteK = 'H' and siteL = 'H'
                        index0 = sites_cumsum[k] + move_pos_asso[k, 0]
                        indexf = sites_cumsum[l] + move_pos_asso[l, 0]
                        epsAB_kl[index0, indexf] = epsAB
                        epsAB_kl[indexf, index0] = epsAB
                        kAB_kl[index0, indexf] = kAB
                        kAB_kl[indexf, index0] = kAB

                        indexAB_id1.append(molecule_id_index_sites[index0])
                        indexAB_id2.append(molecule_id_index_sites[indexf])
                        indexABij1.append(index0)
                        indexABij2.append(indexf)

    indexABij1 = np.hstack([indexABij1])
    indexABij2 = np.hstack([indexABij2])

    indexAB_id1 = np.hstack([indexAB_id1])
    indexAB_id2 = np.hstack([indexAB_id2])

    # make sure association indeces are integers
    indexABij1 = indexABij1.astype(int)
    indexABij2 = indexABij2.astype(int)
    indexAB_id1 = indexAB_id1.astype(int)
    indexAB_id2 = indexAB_id2.astype(int)

    indexAB_id = (indexAB_id1, indexAB_id2)
    indexABij = (indexABij1, indexABij2)

    out = [kAB_kl, epsAB_kl, sites_asso, group_asso_index, nsites,
           molecule_id_index_sites, indexAB_id, indexABij, subgroup_id_asso,
           molecule_id_index_asso, sites_cumsum]
    return out



from __future__ import division, print_function, absolute_import

from .path_sk import ten_beta0_sk
from .reference_component import ten_beta0_reference

from .sgt_beta0 import sgt_mix_beta0
from .sgtpure import sgt_pure
from .linear_spot import sgt_linear, sgt_spot

from .coloc_z import sgt_mix, sgt_zfixed
from .coloc_z_ds import msgt_mix

from .tensionresult import TensionResult



"""
SGTPy.equilibrium: phase equilibria with Python
=======================================================

Obtained from Phasepy's implementation

Functions
---------
bubbleTy : bubble point P, x -> T, y
bubblePy : bubble point T, x -> P, y
dewTx : dew point P, y -> T, x
dewTy : dew point T, y -> P, x
flash : istohermal isobaric two phase flash z, T, P -> x,y,beta
lle : liquid liquid equilibrium  z, T, P -> x, w, beta
lle_init : finds initial guess for ell
multiflash : multiflash algorithm that checks stability of the phases
vlleb : heteroazetropic calculation (VLLE) for binary mixtures
vlle : heteroazetropic calculation (VLLE) for multicomponent mixtures

tpd : Michelsen tpd function
tpd_mim : finds a minimum of tpd function given a initial guess
tpd_minimas : tries to find n minimas of tpd function

"""

from __future__ import division, print_function, absolute_import

# __all__ = [s for s in dir() if not s.startswith('_')]

from .bubble import bubblePy, bubbleTy
from .dew import dewPx, dewTx
from .flash import flash
from .multiflash import multiflash
from .hazt import haz, vlle
from .hazb import vlleb
from .stability import tpd_min, tpd_minimas, lle_init, gmix
from .lle import lle



from __future__ import division, print_function, absolute_import
import numpy as np


# Hard sphere Eq 11
def ahs(eta):
    a = (4*eta - 3*eta**2)/(1-eta)**2
    return a


def dahs_deta(eta):
    a = (4*eta - 3*eta**2)/(1-eta)**2
    da = 2*(-2+eta)/(-1+eta)**3
    return np.array([a, da])


def d2ahs_deta(eta):
    a = (4*eta - 3*eta**2)/(1-eta)**2
    da = 2*(-2+eta)/(-1+eta)**3
    d2a = (10-4*eta)/(-1+eta)**4
    return np.array([a, da, d2a])



from __future__ import division, print_function, absolute_import
import numpy as np


# Equation 65
def g2MCA(rho, suma_g2, da2m_new_deta, khs, eps, cte_g2s, deta_drho):

    da2 = da2m_new_deta * deta_drho
    g2 = (3.*da2 - eps*khs*suma_g2/rho) * cte_g2s

    return g2


def dg2MCA_drho(rho, suma_g2, d2a2m_new_drho, dkhs_drho, eps, cte_g2s):

    khs, dkhs = dkhs_drho
    d2a2 = d2a2m_new_drho
    suma1 = suma_g2 * eps

    dg2 = 3*d2a2 - suma1 * khs/rho
    dg2[1] += suma1[0] * khs/rho**2 - suma1[0] * dkhs / rho
    dg2 *= cte_g2s

    return dg2


def d2g2MCA_drho(rho, suma_g2, d3a2m_new_drho, dkhs_drho, eps, cte_g2s):

    khs, dkhs, d2khs = dkhs_drho

    d3a2 = d3a2m_new_drho

    suma1 = suma_g2 * eps
    sum1, dsum1, d2sum1 = suma1

    d2g2 = 3*d3a2 - suma1 * khs/rho
    d2g2[1] += sum1 * khs/rho**2 - sum1 * dkhs / rho
    d2g2[2] += -2*khs*sum1/rho**3 + 2*sum1*dkhs/rho**2 + 2*khs*dsum1/rho**2
    d2g2[2] += -2*dkhs*dsum1/rho - sum1*d2khs/rho
    d2g2 *= cte_g2s

    return d2g2


phi70 = 10.
phi71 = 10.
phi72 = 0.57
phi73 = -6.7
phi74 = -8


# Equation 63
def gammac(x0, nsigma, alpha, tetha):
    gc = phi70*(-np.tanh(phi71*(phi72-alpha))+1)
    gc *= nsigma*tetha*np.exp(phi73*nsigma + phi74*nsigma**2)
    return gc


def dgammac_deta(x03, nsigma, alpha, tetha):

    cte = phi70*(-np.tanh(phi71*(phi72-alpha))+1)*tetha
    g = cte * np.exp(phi73*nsigma + phi74*nsigma**2)

    dg = np.array([g, g])
    dg[0] *= nsigma

    dg[1] *= (1. + nsigma*(phi73 + 2*phi74*nsigma))
    dg[1] *= x03

    return dg


def d2gammac_deta(x03, nsigma, alpha, tetha):

    cte = phi70*(-np.tanh(phi71*(phi72-alpha))+1)*tetha
    g = cte * np.exp(phi73*nsigma+phi74*nsigma**2)

    dg = np.array([g, g, g])
    dg[0] *= nsigma

    dg[1] *= (1. + nsigma*(phi73 + 2*phi74*nsigma))
    dg[1] *= x03

    aux1 = phi73**2 + 6*phi74+4*phi74*nsigma*(phi73 + phi74*nsigma)
    dg[2] *= (2*phi73 + nsigma * aux1) * x03**2


    return dg



from __future__ import division, print_function, absolute_import
import numpy as np
from .coloc_cy import jcobi_roots, colocAB, colocA, colocB


def gauss(n):
    roots, dif1 = jcobi_roots(n, N0=0, N1=0, Al=0, Be=0)
    ax = 1
    ax /= roots
    ax /= (1-roots)
    vect = ax/dif1**2
    vect /= np.sum(vect)
    return roots, vect


def lobatto(n):
    roots, dif1 = jcobi_roots(n - 2, N0=1, N1=1, Al=1, Be=1)
    s0 = 2/(n-1)/n
    vect = s0/dif1**2
    vect /= np.sum(vect)
    return roots, vect


def gdem(X, X1, X2, X3):
    dX2 = X - X3
    dX1 = X - X2
    dX = X - X1
    b01 = dX.dot(dX1)
    b02 = dX.dot(dX2)
    b12 = dX1.dot(dX2)
    b11 = dX1.dot(dX1)
    b22 = dX2.dot(dX2)
    den = b11*b22-b12**2
    with np.errstate(divide='ignore', invalid='ignore'):
        mu1 = (b02*b12 - b01*b22)/den
        mu2 = (b01*b12 - b02*b11)/den
        dacc = (dX - mu2*dX1)/(1+mu1+mu2)
    return np.nan_to_num(dacc)


__all__ = ['gauss', 'lobatto', 'colocAB', 'colocA', 'colocB', 'gdem']



from __future__ import division, print_function, absolute_import
import numpy as np


phi70 = 10.
phi71 = 10.
phi72 = 0.57
phi73 = -6.7
phi74 = -8


# Equation A37
def gammac(xhixm, alpha, tetha):

    gc = phi70*(-np.tanh(phi71 * (phi72-alpha)) + 1)
    gc *= xhixm*tetha*np.exp(phi73*xhixm + phi74*xhixm**2)

    return gc


def dgammac_dxhi00(xhim, dxhim_dxhi00, alpha, tetha):

    cte = phi70*(-np.tanh(phi71*(phi72-alpha))+1)*tetha
    g = cte*np.exp(phi73*xhim + phi74 * xhim**2)

    dg = np.array([g, g])
    dg[0] *= xhim
    dg[1] *= (1. + xhim*(phi73+2*phi74*xhim))
    dg[1] *= dxhim_dxhi00

    return dg


def d2gammac_dxhi00(xhim, dxhim_dxhi00, alpha, tetha):

    cte = phi70*(-np.tanh(phi71*(phi72-alpha)) + 1) * tetha
    g = cte*np.exp(phi73*xhim + phi74*xhim**2)

    dg = np.array([g, g, g])
    dg[0] *= xhim

    dg[1] *= (1. + xhim*(phi73 + 2*phi74*xhim))
    dg[1] *= dxhim_dxhi00

    aux1 = phi73**2 + 6*phi74+4*phi74*xhim*(phi73 + phi74*xhim)
    dg[2] *= (2*phi73 + xhim * aux1) * dxhim_dxhi00**2

    return dg


def dgammac_dx(xhim, dxhim_dx, alpha, tetha):

    cte = phi70*(-np.tanh(phi71*(phi72-alpha))+1)*tetha
    g = cte*np.exp(phi73*xhim + phi74*xhim**2)
    gc = g * xhim

    g *= 1. + xhim*(phi73 + 2*phi74*xhim)

    dg = np.multiply.outer(dxhim_dx, g)
    return gc, dg


def dgammac_dxxhi(xhim, dxhim_dxhi00, dxhim_dx, alpha, tetha):

    cte = phi70*(-np.tanh(phi71*(phi72-alpha))+1)*tetha
    g = cte * np.exp(phi73*xhim + phi74*xhim**2)
    gc = np.array([g, g])
    gc[0] *= xhim
    gc[1] *= (1. + xhim*(phi73 + 2*phi74*xhim))
    gc[1] *= dxhim_dxhi00

    g *= 1. + xhim*(phi73 + 2*phi74*xhim)

    dg = np.multiply.outer(dxhim_dx, g)
    return gc, dg



from __future__ import division, print_function, absolute_import
import numpy as np


h = 6.626070150e-34  # J s
me = 9.10938291e-31  # 1/Kg


# ideal contribution Eq 68
def aideal(x, rho, beta):
    broglie_vol = h / np.sqrt(2*np.pi * me / beta)
    with np.errstate(divide='ignore', invalid='ignore'):
        logx = np.nan_to_num(np.log(x))
        a = np.dot(x, logx)
        a += np.log(rho * broglie_vol**3)
        a -= 1.
    return a


def daideal_drho(x, rho, beta):
    broglie_vol = h / np.sqrt(2*np.pi * me / beta)
    with np.errstate(divide='ignore', invalid='ignore'):
        logx = np.nan_to_num(np.log(x))
        a = np.dot(x, logx)
        a += np.log(rho * broglie_vol**3) - 1
        da = 1./rho
    return np.hstack([a, da])


def d2aideal_drho(x, rho, beta):
    broglie_vol = h / np.sqrt(2*np.pi * me / beta)
    with np.errstate(divide='ignore', invalid='ignore'):
        logx = np.nan_to_num(np.log(x))
        a = np.dot(x, logx)
        a += np.log(rho * broglie_vol**3) - 1
        da = 1./rho
        d2a = -1/rho**2
    return np.hstack([a, da, d2a])


def daideal_dx(x, rho, beta):
    broglie_vol = h / np.sqrt(2*np.pi * me / beta)
    with np.errstate(divide='ignore', invalid='ignore'):
        logx = np.nan_to_num(np.log(x))
        a = np.dot(x, logx)
        a += np.log(rho * broglie_vol**3)
        a -= 1.
        da = logx + 1.
    return a, da


def daideal_dx_drho(x, rho, beta):
    with np.errstate(divide='ignore', invalid='ignore'):
        logx = np.nan_to_num(np.log(x))
        broglie_vol = h / np.sqrt(2*np.pi * me / beta)
        a = np.dot(x, logx)
        a += np.log(rho * broglie_vol**3)
        a -= 1.
        da = 1./rho

        dax = logx + 1.
    return np.hstack([a, da]), dax



from __future__ import division, print_function, absolute_import
import numpy as np
from .monomer_aux import Xi, dXi, d2Xi
from .monomer_aux import dkHS, d2kHS, d3kHS

from .a1sB_monomer import da1B_eval, d2a1B_eval, d3a1B_eval

from .aHS_monomer import ahs, dahs_deta, d2ahs_deta
from .a2m_monomer import a2m, da2m_deta, d2a2m_deta
from .a2m_monomer import da2m_new_deta, d2a2m_new_deta, d3a2m_new_deta
from .a3m_monomer import a3m, da3m_deta, d2a3m_deta

from .gdHS_chain import gdHS, dgdHS_drho, d2gdHS_drho
from .g1sigma_chain import g1sigma, dg1sigma_drho, d2g1sigma_drho
from .g2sigma_chain import g2MCA, dg2MCA_drho, d2g2MCA_drho
from .g2sigma_chain import gammac, dgammac_deta, d2gammac_deta
from .lngmie_chain import lngmie, dlngmie_drho, d2lngmie_drho

from .association_aux import Xass_solver, Iab, dIab_drho
from .association_aux import d2Iab_drho, dXass_drho, d2Xass_drho

from .polarGV import Apol, dApol_drho, d2Apol_drho


def ares(self, rho, temp_aux, Xass0=None):
    beta, beta2, beta3, dia, dia3, x0, x03, x0_a1, x0_a2 = temp_aux[:9]
    x0_a12, x0_a22, I_lambdas, J_lambdas, beps, beps2, tetha = temp_aux[9:16]
    x0_vector, cte_g1s, cte_g2s = temp_aux[16:19]

    eta, deta = self.eta_bh(rho, dia3)
    nsigma = self.eta_sigma(rho)
    # Parameters needed for evaluating the helmothlz contributions
    a1sb_a1, a1sb_a2 = da1B_eval(eta, I_lambdas, J_lambdas, self.lambdas,
                                 self.cctes, self.eps)
    dkhs = dkHS(eta)
    xi = Xi(x03, nsigma, self.f1, self.f2, self.f3)

    cte_a2m = self.cte_a2m
    eps3 = self.eps3

    # Monomer contribution
    ahs_eval = ahs(eta)
    a1m_eval = self.c*np.matmul(a1sb_a1, x0_a1)
    suma_a2 = np.matmul(a1sb_a2, x0_a2)
    a2m_eval = a2m(suma_a2[0], dkhs[0], xi, cte_a2m)
    a3m_eval = a3m(x03, nsigma, eps3, self.f4, self.f5, self.f6)

    a_mono = ahs_eval + beta*a1m_eval[0] + beta2*a2m_eval + beta3*a3m_eval
    a_mono *= self.ms

    # chain contribution calculation
    ghs = gdHS(x0_vector, eta)

    # g1sigma
    suma_g1 = self.c * np.dot(a1sb_a1[0], x0_a12)
    g1s = g1sigma(rho, suma_g1, a1m_eval[1], deta, cte_g1s)

    # g2sigma
    gc = gammac(x0, nsigma, self.alpha, tetha)
    da2m_new = da2m_new_deta(suma_a2, dkhs, cte_a2m)

    suma_g2 = self.c2*np.dot(a1sb_a2[0],  x0_a22)
    g2m = g2MCA(rho, suma_g2, da2m_new, dkhs[0], self.eps, cte_g2s, deta)
    g2s = (1+gc)*g2m

    lng = lngmie(ghs, g1s, g2s, beps, beps2)
    a_chain = - (self.ms - 1. + self.ring*eta)*lng

    # Total Helmolthz
    a = a_mono + a_chain

    if self.assoc_bool:

        if Xass0 is None:
            Xass = 0.2 * np.ones(self.nsites)
        else:
            Xass = 1. * Xass0
        Fab = temp_aux[19]
        Kab = temp_aux[20]
        iab = Iab(Kab, eta)
        Dab = self.sigma3 * Fab * iab
        Dabij = np.zeros([self.nsites, self.nsites])
        Dabij[self.indexabij] = Dab
        KIJ = rho * (self.DIJ*Dabij)
        Xass = Xass_solver(self.nsites, KIJ, self.diagasso, Xass0)
        a += np.dot(self.S, (np.log(Xass) - Xass/2 + 1/2))
    else:
        Xass = None

    if self.polar_bool:
        epsa = temp_aux[21]
        apolar = Apol(rho, eta, epsa, self.anij, self.bnij,
                      self.cnijk, self.mupolad2, self.npol, self.sigma3)
        a += apolar

    return a, Xass


def dares_drho(self, rho, temp_aux, Xass0=None):
    beta, beta2, beta3, dia, dia3, x0, x03, x0_a1, x0_a2 = temp_aux[:9]
    x0_a12, x0_a22, I_lambdas, J_lambdas, beps, beps2, tetha = temp_aux[9:16]
    x0_vector, cte_g1s, cte_g2s = temp_aux[16:19]

    eta, deta = self.eta_bh(rho, dia3)
    nsigma = self.eta_sigma(rho)

    drho = np.array([1., deta, deta**2])

    # Parameters needed for evaluating the helmothlz contributions
    a1sb_a1, a1sb_a2 = d2a1B_eval(eta, I_lambdas, J_lambdas, self.lambdas,
                                  self.cctes, self.eps)
    dkhs = d2kHS(eta)
    dxi = dXi(x03, nsigma, self.f1, self.f2, self.f3)

    cte_a2m = self.cte_a2m
    eps3 = self.eps3

    # monomer evaluation
    ahs_eval = dahs_deta(eta)
    a1m_eval = self.c*np.matmul(a1sb_a1, x0_a1)
    suma_a2 = np.matmul(a1sb_a2, x0_a2)
    a2m_eval = da2m_deta(suma_a2[:2], dkhs[:2], dxi, cte_a2m)
    a3m_eval = da3m_deta(x03, nsigma, eps3, self.f4, self.f5, self.f6)

    a_mono = ahs_eval + beta*a1m_eval[:2] + beta2*a2m_eval + beta3*a3m_eval
    a_mono *= self.ms * drho[:2]

    # chain contribution calculation
    dghs = dgdHS_drho(x0_vector, eta, drho)

    # g1sigma
    suma_g1 = self.c * np.dot(a1sb_a1[:2], x0_a12)
    suma_g1 *= drho[:2]
    d2a1m_drho = a1m_eval[1:]*drho[1:]
    dg1s = dg1sigma_drho(rho, suma_g1, d2a1m_drho, cte_g1s)

    # g2sigma
    dgc = dgammac_deta(x03, nsigma, self.alpha, tetha)
    dgc *= drho[:2]

    da2m_new = d2a2m_new_deta(suma_a2, dkhs, cte_a2m)
    da2m_new_drho = da2m_new*drho[1:]

    suma_g2 = self.c2*np.dot(a1sb_a2[:2],  x0_a22)
    suma_g2 *= drho[:2]

    dkhs_drho = dkhs[:2]*drho[:2]

    dg2m = dg2MCA_drho(rho, suma_g2, da2m_new_drho, dkhs_drho, self.eps,
                       cte_g2s)

    dg2s = dg2m * (1. + dgc[0])
    dg2s[1] += dgc[1] * dg2m[0]

    dlng = dlngmie_drho(dghs, dg1s, dg2s, beps, beps2)
    a_chain = - (self.ms - 1. + self.ring*eta)*dlng

    # Total helmolthz
    a = a_mono + a_chain

    if self.assoc_bool:
        if Xass0 is None:
            Xass = 0.2 * np.ones(self.nsites)
        else:
            Xass = 1. * Xass0
        Fab = temp_aux[19]
        Kab = temp_aux[20]
        iab, diab = dIab_drho(Kab, eta, deta)
        # Fab = np.exp(beta * self.eABij) - 1
        Dab = self.sigma3 * Fab * iab
        dDab = self.sigma3 * Fab * diab
        Dabij = np.zeros([self.nsites, self.nsites])
        dDabij_drho = np.zeros([self.nsites, self.nsites])

        Dabij[self.indexabij] = Dab
        dDabij_drho[self.indexabij] = dDab
        KIJ = rho * (self.DIJ*Dabij)
        Xass = Xass_solver(self.nsites, KIJ, self.diagasso, Xass0)

        CIJ = rho * np.tile(Xass**2, (self.nsites, 1)).T * Dabij * self.DIJ
        CIJ[self.diagasso] += 1.
        dXass = dXass_drho(rho, Xass, self.DIJ, Dabij, dDabij_drho, CIJ)
        a[0] += np.dot(self.S, (np.log(Xass) - Xass/2 + 1/2))
        a[1] += np.dot(self.S, (1/Xass - 1/2) * dXass)
    else:
        Xass = None

    if self.polar_bool:
        epsa = temp_aux[21]
        dapolar = dApol_drho(rho, eta, deta, epsa, self.anij,
                             self.bnij, self.cnijk, self.mupolad2,
                             self.npol, self.sigma3)
        a += dapolar

    return a, Xass


def d2ares_drho(self, rho, temp_aux, Xass0=None):

    beta, beta2, beta3, dia, dia3, x0, x03, x0_a1, x0_a2 = temp_aux[:9]
    x0_a12, x0_a22, I_lambdas, J_lambdas, beps, beps2, tetha = temp_aux[9:16]
    x0_vector, cte_g1s, cte_g2s = temp_aux[16:19]

    eta, deta = self.eta_bh(rho, dia3)
    nsigma = self.eta_sigma(rho)

    drho = np.array([1., deta, deta**2, deta**3])

    # Parameters needed for evaluating the helmothlz contributions
    a1sb_a1, a1sb_a2 = d3a1B_eval(eta, I_lambdas, J_lambdas, self.lambdas,
                                  self.cctes, self.eps)

    dkhs = d3kHS(eta)
    dxi = d2Xi(x03, nsigma, self.f1, self.f2, self.f3)

    cte_a2m = self.cte_a2m
    eps3 = self.eps3

    # monomer evaluation
    ahs_eval = d2ahs_deta(eta)
    a1m_eval = self.c*np.matmul(a1sb_a1, x0_a1)
    suma_a2 = np.matmul(a1sb_a2, x0_a2)
    a2m_eval = d2a2m_deta(suma_a2[:3], dkhs[:3], dxi, cte_a2m)
    a3m_eval = d2a3m_deta(x03, nsigma, eps3, self.f4, self.f5, self.f6)

    a_mono = ahs_eval + beta*a1m_eval[:3] + beta2*a2m_eval + beta3*a3m_eval
    a_mono *= self.ms * drho[:3]

    # chain contribution calculation
    dghs = d2gdHS_drho(x0_vector, eta, drho)

    # g1sigma
    suma_g1 = self.c * np.dot(a1sb_a1[:3], x0_a12)
    suma_g1 *= drho[:3]
    d3a1m_drho = a1m_eval[1:]*drho[1:]
    dg1s = d2g1sigma_drho(rho, suma_g1, d3a1m_drho, cte_g1s)

    # g2sigma
    dgc = d2gammac_deta(x03, nsigma, self.alpha, tetha)
    dgc *= drho[:3]

    da2m_new = d3a2m_new_deta(suma_a2, dkhs, cte_a2m)
    da2m_new_drho = da2m_new*drho[1:]

    suma_g2 = self.c2*np.dot(a1sb_a2[:3],  x0_a22)
    suma_g2 *= drho[:3]

    dkhs_drho = dkhs[:3]*drho[:3]

    dg2m = d2g2MCA_drho(rho, suma_g2, da2m_new_drho, dkhs_drho, self.eps,
                        cte_g2s)

    dg2s = dg2m * (1. + dgc[0])
    dg2s[1] += dgc[1] * dg2m[0]
    dg2s[2] += dgc[2]*dg2m[0] + 2.*dgc[1]*dg2m[1]

    d2lng = d2lngmie_drho(dghs, dg1s, dg2s, beps, beps2)
    a_chain = - (self.ms - 1. + self.ring*eta)*d2lng

    # Total Helmolthz
    a = a_mono + a_chain

    if self.assoc_bool:
        if Xass0 is None:
            Xass = 0.2 * np.ones(self.nsites)
        else:
            Xass = 1. * Xass0
        Fab = temp_aux[19]
        Kab = temp_aux[20]
        iab, diab, d2iab = d2Iab_drho(Kab, eta, deta)

        # Fab = np.exp(beta * self.eABij) - 1.
        Dab = self.sigma3 * Fab * iab
        dDab = self.sigma3 * Fab * diab
        d2Dab = self.sigma3 * Fab * d2iab

        Dabij = np.zeros([self.nsites, self.nsites])
        dDabij_drho = np.zeros([self.nsites, self.nsites])
        d2Dabij_drho = np.zeros([self.nsites, self.nsites])
        Dabij[self.indexabij] = Dab
        dDabij_drho[self.indexabij] = dDab
        d2Dabij_drho[self.indexabij] = d2Dab

        KIJ = rho * (self.DIJ*Dabij)
        Xass = Xass_solver(self.nsites, KIJ, self.diagasso, Xass0)

        CIJ = rho * np.tile(Xass**2, (self.nsites, 1)).T * Dabij * self.DIJ
        CIJ[self.diagasso] += 1.
        dXass = dXass_drho(rho, Xass, self.DIJ, Dabij, dDabij_drho, CIJ)
        d2Xass = d2Xass_drho(rho, Xass, dXass, self.DIJ, Dabij, dDabij_drho,
                             d2Dabij_drho, CIJ)
        a[0] += np.dot(self.S, (np.log(Xass) - Xass/2 + 1/2))
        a[1] += np.dot(self.S, (1/Xass - 1/2) * dXass)
        a[2] += np.dot(self.S, - (dXass/Xass)**2 + d2Xass * (1/Xass - 1/2))
    else:
        Xass = None

    if self.polar_bool:
        epsa = temp_aux[21]
        dapolar = d2Apol_drho(rho, eta, deta, epsa, self.anij,
                              self.bnij, self.cnijk, self.mupolad2,
                              self.npol, self.sigma3)

        a += dapolar

    return a, Xass



from __future__ import division, print_function, absolute_import
import numpy as np


def association_config(eos):

    types = np.array(['B', 'P', 'N'])
    nozero = np.nonzero(eos.sites)
    types = types[nozero]
    ntypes = np.asarray(eos.sites)[nozero]
    nsites = len(types)
    S = np.array(eos.sites)
    S = S[S != 0]
    DIJ = np.zeros([nsites, nsites])
    int_i = []
    int_j = []
    for i in range(nsites):
        for j in range(nsites):
            bool1 = types[i] == 'B'
            bool2 = types[i] == 'P' and (types[j] == 'N' or types[j] == 'B')
            bool3 = types[i] == 'N' and (types[j] == 'P' or types[j] == 'B')
            if bool1 or bool2 or bool3:
                DIJ[i, j] = ntypes[j]
                int_i.append(i)
                int_j.append(j)

    indexabij = (int_i, int_j)
    diagasso = np.diag_indices(nsites)

    return S, DIJ, indexabij, nsites, diagasso


def Xass_solver(nsites, KIJ, diagasso, Xass0=None):

    if Xass0 is None:
        Xass = 0.2 * np.ones(nsites)
    else:
        Xass = 1. * Xass0
    omega = 0.2

    for i in range(5):
        fo = 1. / (1. + KIJ@Xass)
        dXass = (1 - omega) * (fo - Xass)
        Xass += dXass

    KIJXass = KIJ@Xass
    dQ = (1./Xass - 1.) - KIJXass
    HIJ = - 1. * KIJ
    HIJ[diagasso] -= (1. + KIJXass)/Xass
    for i in range(15):
        dXass = np.linalg.solve(HIJ, -dQ)
        Xnew = Xass + dXass

        is_nan = np.isnan(Xnew)
        Xnew[is_nan] = 0.2

        Xnew_neg = Xnew < 0.
        Xnew[Xnew_neg] = 0.2*Xass[Xnew_neg]

        Xnew_big = Xnew > 1.
        Xnew[Xnew_big] = 1.

        Xass = Xnew
        KIJXass = KIJ@Xass
        dQ = (1./Xass - 1.) - KIJXass
        sucess = np.linalg.norm(dQ) < 1e-9
        if sucess:
            break
        HIJ = - 1. * KIJ
        HIJ[diagasso] -= (1. + KIJXass)/Xass

    return Xass


def Iab(Kab, eta):

    gdhs = (1 - eta/2) / (1 - eta)**3
    Iab = Kab * gdhs

    return Iab


def dIab_drho(Kab, eta, deta_drho):

    eta_1 = 1-eta
    eta_13 = eta_1**3
    eta_14 = eta_13*eta_1

    gdhs = (1 - eta/2) / eta_13
    dgdhs = (2.5 - eta) * deta_drho / eta_14

    Iab = Kab * gdhs
    dIab = Kab * dgdhs
    return Iab, dIab


def d2Iab_drho(Kab, eta, deta_drho):
    eta_1 = 1-eta
    eta_13 = eta_1**3
    eta_14 = eta_13*eta_1
    eta_15 = eta_14*eta_1

    gdhs = (1 - eta/2) / eta_13
    dgdhs = (2.5 - eta) * deta_drho / eta_14
    d2gdhs = 3 * (3 - eta) * deta_drho**2 / eta_15

    Iab = Kab * gdhs
    dIab = Kab * dgdhs
    d2Iab = Kab * d2gdhs
    return Iab, dIab, d2Iab


def dXass_drho(rho, Xass, DIJ, Dabij, dDabij_drho, CIJ):
    brho = -(DIJ*(Dabij + rho * dDabij_drho))@Xass
    brho *= Xass**2
    dXass = np.linalg.solve(CIJ, brho)
    return dXass


def d2Xass_drho(rho, Xass, dXass_drho, DIJ, Dabij, dDabij_drho, d2Dabij_drho,
                CIJ):

    b2rho = np.sum(DIJ*(Xass*d2Dabij_drho + 2*dXass_drho*dDabij_drho), axis=1)
    b2rho *= - rho
    b2rho += 2 * (1/Xass - 1) / (rho**2)
    b2rho *= Xass**2
    b2rho += 2 * dXass_drho / (rho)
    b2rho += 2 * dXass_drho**2 / (Xass)

    d2Xass_drho = np.linalg.solve(CIJ, b2rho)
    return d2Xass_drho


def association_solver(self, rhom, temp_aux, Xass0=None):
    if Xass0 is None:
        Xass = 0.2 * np.ones(self.nsites)
    else:
        Xass = 1. * Xass0

    dia3 = temp_aux[4]
    eta, deta = self.eta_bh(rhom, dia3)

    Fab = temp_aux[19]
    Kab = temp_aux[20]
    iab = Iab(Kab, eta)
    Dab = self.sigma3 * Fab * iab
    Dabij = np.zeros([self.nsites, self.nsites])
    Dabij[self.indexabij] = Dab
    KIJ = rhom * (self.DIJ*Dabij)
    Xass = Xass_solver(self.nsites, KIJ, self.diagasso, Xass0)
    return Xass


def association_check(self, rhom, temp_aux, Xass):

    dia3 = temp_aux[4]
    eta, deta = self.eta_bh(rhom, dia3)

    Fab = temp_aux[19]
    Kab = temp_aux[20]
    iab = Iab(Kab, eta)
    Dab = self.sigma3 * Fab * iab
    Dabij = np.zeros([self.nsites, self.nsites])
    Dabij[self.indexabij] = Dab
    KIJ = rhom * (self.DIJ*Dabij)
    fo = Xass - 1. / (1. + KIJ@Xass)
    return fo
